# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Pack repository objects."""

import hashlib
import logging
import sys
import time

logger = logging.getLogger(__name__)

from . import btree_index
from . import pack as _mod_pack
from .errors import BzrCheckError, BzrFormatsError
from .transport import TransportNoSuchFile


class RetryWithNewPacks(BzrFormatsError):
    """Raised when we realize that the packs on disk have changed.

    This is meant as more of a signaling exception, to trap between where a
    local error occurred and the code that can actually handle the error and
    code that can retry appropriately.
    """

    internal_error = True

    _fmt = (
        "Pack files have changed, reload and retry. context: %(context)s %(orig_error)s"
    )

    def __init__(self, context, reload_occurred, exc_info):
        """Create a new RetryWithNewPacks error.

        :param reload_occurred: Set to True if we know that the packs have
            already been reloaded, and we are failing because of an in-memory
            cache miss. If set to True then we will ignore if a reload says
            nothing has changed, because we assume it has already reloaded. If
            False, then a reload with nothing changed will force an error.
        :param exc_info: The original exception traceback, so if there is a
            problem we can raise the original error (value from sys.exc_info())
        """
        BzrFormatsError.__init__(self)
        self.context = context
        self.reload_occurred = reload_occurred
        self.exc_info = exc_info
        self.orig_error = exc_info[1]
        # TODO: The global error handler should probably treat this by
        #       raising/printing the original exception with a bit about
        #       RetryWithNewPacks also not being caught


class _DirectPackAccess:
    """Access to data in one or more packs with less translation."""

    def __init__(self, index_to_packs, reload_func=None, flush_func=None):
        """Create a _DirectPackAccess object.

        :param index_to_packs: A dict mapping index objects to the transport
            and file names for obtaining data.
        :param reload_func: A function to call if we determine that the pack
            files have moved and we need to reload our caches. See
            breezy.repo_fmt.pack_repo.AggregateIndex for more details.
        """
        self._container_writer = None
        self._write_index = None
        self._indices = index_to_packs
        self._reload_func = reload_func
        self._flush_func = flush_func

    def add_raw_record(self, key, size, raw_data):
        """Add raw knit bytes to a storage area.

        The data is spooled to the container writer in one bytes-record per
        raw data item.

        :param key: key of the data segment
        :param size: length of the data segment
        :param raw_data: A bytestring containing the data.
        :return: An opaque index memo For _DirectPackAccess the memo is
            (index, pos, length), where the index field is the write_index
            object supplied to the PackAccess object.
        """
        p_offset, p_length = self._container_writer.add_bytes_record(raw_data, size, [])
        return (self._write_index, p_offset, p_length)

    def add_raw_records(self, key_sizes, raw_data):
        """Add raw knit bytes to a storage area.

        The data is spooled to the container writer in one bytes-record per
        raw data item.

        :param sizes: An iterable of tuples containing the key and size of each
            raw data segment.
        :param raw_data: A bytestring containing the data.
        :return: A list of memos to retrieve the record later. Each memo is an
            opaque index memo. For _DirectPackAccess the memo is (index, pos,
            length), where the index field is the write_index object supplied
            to the PackAccess object.
        """
        raw_data = b"".join(raw_data)
        if not isinstance(raw_data, bytes):
            raise AssertionError(f"data must be plain bytes was {type(raw_data)}")
        result = []
        offset = 0
        for key, size in key_sizes:
            result.append(
                self.add_raw_record(key, size, [raw_data[offset : offset + size]])
            )
            offset += size
        return result

    def flush(self):
        """Flush pending writes on this access object.

        This will flush any buffered writes to a NewPack.
        """
        if self._flush_func is not None:
            self._flush_func()

    def get_raw_records(self, memos_for_retrieval):
        """Get the raw bytes for a records.

        :param memos_for_retrieval: An iterable containing the (index, pos,
            length) memo for retrieving the bytes. The Pack access method
            looks up the pack to use for a given record in its index_to_pack
            map.
        :return: An iterator over the bytes of the records.
        """
        # first pass, group into same-index requests
        request_lists = []
        current_index = None
        for index, offset, length in memos_for_retrieval:
            if current_index == index:
                current_list.append((offset, length))
            else:
                if current_index is not None:
                    request_lists.append((current_index, current_list))
                current_index = index
                current_list = [(offset, length)]
        # handle the last entry
        if current_index is not None:
            request_lists.append((current_index, current_list))
        for index, offsets in request_lists:
            try:
                transport, path = self._indices[index]
            except KeyError as e:
                # A KeyError here indicates that someone has triggered an index
                # reload, and this index has gone missing, we need to start
                # over.
                if self._reload_func is None:
                    # If we don't have a _reload_func there is nothing that can
                    # be done
                    raise
                raise RetryWithNewPacks(
                    index, reload_occurred=True, exc_info=sys.exc_info()
                ) from e
            try:
                reader = _mod_pack.make_readv_reader(transport, path, offsets)
                for _names, read_func in reader.iter_records():
                    yield read_func(None)
            except TransportNoSuchFile as e:
                # A NoSuchFile error indicates that a pack file has gone
                # missing on disk, we need to trigger a reload, and start over.
                if self._reload_func is None:
                    raise
                raise RetryWithNewPacks(
                    transport.abspath(path),
                    reload_occurred=False,
                    exc_info=sys.exc_info(),
                ) from e

    def set_writer(self, writer, index, transport_packname):
        """Set a writer to use for adding data."""
        if index is not None:
            self._indices[index] = transport_packname
        self._container_writer = writer
        self._write_index = index

    def reload_or_raise(self, retry_exc):
        """Try calling the reload function, or re-raise the original exception.

        This should be called after _DirectPackAccess raises a
        RetryWithNewPacks exception. This function will handle the common logic
        of determining when the error is fatal versus being temporary.
        It will also make sure that the original exception is raised, rather
        than the RetryWithNewPacks exception.

        If this function returns, then the calling function should retry
        whatever operation was being performed. Otherwise an exception will
        be raised.

        :param retry_exc: A RetryWithNewPacks exception.
        """
        is_error = False
        if self._reload_func is None:
            is_error = True
        elif not self._reload_func():
            # The reload claimed that nothing changed
            if not retry_exc.reload_occurred:
                # If there wasn't an earlier reload, then we really were
                # expecting to find changes. We didn't find them, so this is a
                # hard error
                is_error = True
        if is_error:
            # GZ 2017-03-27: No real reason this needs the original traceback.
            raise retry_exc.exc_info[1]


class Pack:
    """An in memory proxy for a pack and its indices.

    This is a base class that is not directly used, instead the classes
    ExistingPack and NewPack are used.
    """

    # A map of index 'type' to the file extension and position in the
    # index_sizes array.
    index_definitions = {
        "chk": (".cix", 4),
        "revision": (".rix", 0),
        "inventory": (".iix", 1),
        "text": (".tix", 2),
        "signature": (".six", 3),
    }

    def __init__(
        self,
        revision_index,
        inventory_index,
        text_index,
        signature_index,
        chk_index=None,
    ):
        """Create a pack instance.

        :param revision_index: A GraphIndex for determining what revisions are
            present in the Pack and accessing the locations of their texts.
        :param inventory_index: A GraphIndex for determining what inventories are
            present in the Pack and accessing the locations of their
            texts/deltas.
        :param text_index: A GraphIndex for determining what file texts
            are present in the pack and accessing the locations of their
            texts/deltas (via (fileid, revisionid) tuples).
        :param signature_index: A GraphIndex for determining what signatures are
            present in the Pack and accessing the locations of their texts.
        :param chk_index: A GraphIndex for accessing content by CHK, if the
            pack has one.
        """
        self.revision_index = revision_index
        self.inventory_index = inventory_index
        self.text_index = text_index
        self.signature_index = signature_index
        self.chk_index = chk_index

    def access_tuple(self):
        """Return a tuple (transport, name) for the pack content."""
        return self.pack_transport, self.file_name()

    def _check_references(self):
        """Make sure our external references are present.

        Packs are allowed to have deltas whose base is not in the pack, but it
        must be present somewhere in this collection.  It is not allowed to
        have deltas based on a fallback repository.
        (See <https://bugs.launchpad.net/bzr/+bug/288751>)
        """
        missing_items = {}
        for index_name, external_refs, index in [
            (
                "texts",
                self._get_external_refs(self.text_index),
                self._pack_collection.text_index.combined_index,
            ),
            (
                "inventories",
                self._get_external_refs(self.inventory_index),
                self._pack_collection.inventory_index.combined_index,
            ),
        ]:
            missing = external_refs.difference(
                k for (idx, k, v, r) in index.iter_entries(external_refs)
            )
            if missing:
                missing_items[index_name] = sorted(missing)
        if missing_items:
            from pprint import pformat

            raise BzrCheckError(
                f"Newly created pack file {self!r} has delta references to "
                f"items not in its repository:\n{pformat(missing_items)}"
            )

    def file_name(self):
        """Get the file name for the pack on disk."""
        return self.name + ".pack"

    def get_revision_count(self):
        """Return the number of revisions in this pack."""
        return self.revision_index.key_count()

    def index_name(self, index_type, name):
        """Get the disk name of an index type for pack name 'name'."""
        return name + Pack.index_definitions[index_type][0]

    def index_offset(self, index_type):
        """Get the position in a index_size array for a given index type."""
        return Pack.index_definitions[index_type][1]

    def inventory_index_name(self, name):
        """The inv index is the name + .iix."""
        return self.index_name("inventory", name)

    def revision_index_name(self, name):
        """The revision index is the name + .rix."""
        return self.index_name("revision", name)

    def signature_index_name(self, name):
        """The signature index is the name + .six."""
        return self.index_name("signature", name)

    def text_index_name(self, name):
        """The text index is the name + .tix."""
        return self.index_name("text", name)

    def _replace_index_with_readonly(self, index_type):
        unlimited_cache = False
        if index_type == "chk":
            unlimited_cache = True
        index = self.index_class(
            self.index_transport,
            self.index_name(index_type, self.name),
            self.index_sizes[self.index_offset(index_type)],
            unlimited_cache=unlimited_cache,
        )
        if index_type == "chk":
            index._leaf_factory = btree_index._gcchk_factory
        setattr(self, index_type + "_index", index)

    def __lt__(self, other):
        """Compare packs by identity for ordering."""
        if not isinstance(other, Pack):
            raise TypeError(other)
        return id(self) < id(other)

    def __hash__(self):
        """Return hash based on index objects."""
        return hash(
            (
                type(self),
                self.revision_index,
                self.inventory_index,
                self.text_index,
                self.signature_index,
                self.chk_index,
            )
        )


class ExistingPack(Pack):
    """An in memory proxy for an existing .pack and its disk indices."""

    def __init__(
        self,
        pack_transport,
        name,
        revision_index,
        inventory_index,
        text_index,
        signature_index,
        chk_index=None,
    ):
        """Create an ExistingPack object.

        :param pack_transport: The transport where the pack file resides.
        :param name: The name of the pack on disk in the pack_transport.
        """
        Pack.__init__(
            self,
            revision_index,
            inventory_index,
            text_index,
            signature_index,
            chk_index,
        )
        self.name = name
        self.pack_transport = pack_transport
        if None in (
            revision_index,
            inventory_index,
            text_index,
            signature_index,
            name,
            pack_transport,
        ):
            raise AssertionError()

    def __eq__(self, other):
        """Check equality by comparing all attributes."""
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Check inequality."""
        return not self.__eq__(other)

    def __repr__(self):
        """Return string representation."""
        return "<{}.{} object at 0x{:x}, {}, {}".format(
            self.__class__.__module__,
            self.__class__.__name__,
            id(self),
            self.pack_transport,
            self.name,
        )

    def __hash__(self):
        """Return hash based on type and name."""
        return hash((type(self), self.name))


class ResumedPack(ExistingPack):
    """A pack being resumed from an interrupted upload."""

    def __init__(
        self,
        name,
        revision_index,
        inventory_index,
        text_index,
        signature_index,
        upload_transport,
        pack_transport,
        index_transport,
        pack_collection,
        chk_index=None,
    ):
        """Create a ResumedPack object."""
        ExistingPack.__init__(
            self,
            pack_transport,
            name,
            revision_index,
            inventory_index,
            text_index,
            signature_index,
            chk_index=chk_index,
        )
        self.upload_transport = upload_transport
        self.index_transport = index_transport
        self.index_sizes = [None, None, None, None]
        indices = [
            ("revision", revision_index),
            ("inventory", inventory_index),
            ("text", text_index),
            ("signature", signature_index),
        ]
        if chk_index is not None:
            indices.append(("chk", chk_index))
            self.index_sizes.append(None)
        for index_type, index in indices:
            offset = self.index_offset(index_type)
            self.index_sizes[offset] = index._size
        self.index_class = pack_collection._index_class
        self._pack_collection = pack_collection
        self._state = "resumed"
        # XXX: perhaps check that the .pack file exists?

    def access_tuple(self):
        """Return the transport and file name for accessing the pack data."""
        if self._state == "finished":
            return Pack.access_tuple(self)
        elif self._state == "resumed":
            return self.upload_transport, self.file_name()
        else:
            raise AssertionError(self._state)

    def abort(self):
        """Abort the resumed pack, deleting its files."""
        self.upload_transport.delete(self.file_name())
        indices = [
            self.revision_index,
            self.inventory_index,
            self.text_index,
            self.signature_index,
        ]
        if self.chk_index is not None:
            indices.append(self.chk_index)
        for index in indices:
            index._transport.delete(index._name)

    def finish(self):
        """Finish the resumed pack, moving files into place."""
        self._check_references()
        index_types = ["revision", "inventory", "text", "signature"]
        if self.chk_index is not None:
            index_types.append("chk")
        for index_type in index_types:
            old_name = self.index_name(index_type, self.name)
            new_name = "../indices/" + old_name
            self.upload_transport.move(old_name, new_name)
            self._replace_index_with_readonly(index_type)
        new_name = "../packs/" + self.file_name()
        self.upload_transport.move(self.file_name(), new_name)
        self._state = "finished"

    def _get_external_refs(self, index):
        """Return compression parents for this index that are not present.

        This returns any compression parents that are referenced by this index,
        which are not contained *in* this index. They may be present elsewhere.
        """
        return index.external_references(1)


class NewPack(Pack):
    """An in memory proxy for a pack which is being created."""

    def __init__(self, pack_collection, upload_suffix="", file_mode=None):
        """Create a NewPack instance.

        :param pack_collection: A PackCollection into which this is being inserted.
        :param upload_suffix: An optional suffix to be given to any temporary
            files created during the pack creation. e.g '.autopack'
        :param file_mode: Unix permissions for newly created file.
        """
        # The relative locations of the packs are constrained, but all are
        # passed in because the caller has them, so as to avoid object churn.
        index_builder_class = pack_collection._index_builder_class
        if pack_collection.chk_index is not None:
            chk_index = index_builder_class(reference_lists=0)
        else:
            chk_index = None
        Pack.__init__(
            self,
            # Revisions: parents list, no text compression.
            index_builder_class(reference_lists=1),
            # Inventory: We want to map compression only, but currently the
            # knit code hasn't been updated enough to understand that, so we
            # have a regular 2-list index giving parents and compression
            # source.
            index_builder_class(reference_lists=2),
            # Texts: compression and per file graph, for all fileids - so two
            # reference lists and two elements in the key tuple.
            index_builder_class(reference_lists=2, key_elements=2),
            # Signatures: Just blobs to store, no compression, no parents
            # listing.
            index_builder_class(reference_lists=0),
            # CHK based storage - just blobs, no compression or parents.
            chk_index=chk_index,
        )
        self._pack_collection = pack_collection
        # When we make readonly indices, we need this.
        self.index_class = pack_collection._index_class
        # where should the new pack be opened
        self.upload_transport = pack_collection._upload_transport
        # where are indices written out to
        self.index_transport = pack_collection._index_transport
        # where is the pack renamed to when it is finished?
        self.pack_transport = pack_collection._pack_transport
        # What file mode to upload the pack and indices with.
        self._file_mode = file_mode
        # tracks the content written to the .pack file.
        self._hash = hashlib.md5()  # noqa: S324
        # a tuple with the length in bytes of the indices, once the pack
        # is finalised. (rev, inv, text, sigs, chk_if_in_use)
        self.index_sizes = None
        # How much data to cache when writing packs. Note that this is not
        # synchronised with reads, because it's not in the transport layer, so
        # is not safe unless the client knows it won't be reading from the pack
        # under creation.
        self._cache_limit = 0
        # the temporary pack file name.
        from .osutils import rand_chars

        self.random_name = rand_chars(20) + upload_suffix
        # when was this pack started ?
        self.start_time = time.time()
        # open an output stream for the data added to the pack.
        self.write_stream = self.upload_transport.open_write_stream(
            self.random_name, mode=self._file_mode
        )
        logger.debug(
            "%s: create_pack: pack stream open: %s%s t+%6.3fs",
            time.ctime(),
            self.upload_transport.base,
            self.random_name,
            time.time() - self.start_time,
        )
        # A list of byte sequences to be written to the new pack, and the
        # aggregate size of them.  Stored as a list rather than separate
        # variables so that the _write_data closure below can update them.
        self._buffer = [[], 0]
        # create a callable for adding data
        #
        # robertc says- this is a closure rather than a method on the object
        # so that the variables are locals, and faster than accessing object
        # members.

        def _write_data(
            bytes,
            flush=False,
            _buffer=self._buffer,
            _write=self.write_stream.write,
            _update=self._hash.update,
        ):
            _buffer[0].append(bytes)
            _buffer[1] += len(bytes)
            # buffer cap
            if _buffer[1] > self._cache_limit or flush:
                bytes = b"".join(_buffer[0])
                _write(bytes)
                _update(bytes)
                _buffer[:] = [[], 0]

        # expose this on self, for the occasion when clients want to add data.
        self._write_data = _write_data
        # a pack writer object to serialise pack records.
        self._writer = _mod_pack.ContainerWriter(self._write_data)
        self._writer.begin()
        # what state is the pack in? (open, finished, aborted)
        self._state = "open"
        # no name until we finish writing the content
        self.name = None

    def abort(self):
        """Cancel creating this pack."""
        self._state = "aborted"
        self.write_stream.close()
        # Remove the temporary pack file.
        self.upload_transport.delete(self.random_name)
        # The indices have no state on disk.

    def access_tuple(self):
        """Return a tuple (transport, name) for the pack content."""
        if self._state == "finished":
            return Pack.access_tuple(self)
        elif self._state == "open":
            return self.upload_transport, self.random_name
        else:
            raise AssertionError(self._state)

    def data_inserted(self):
        """True if data has been added to this pack."""
        return bool(
            self.get_revision_count()
            or self.inventory_index.key_count()
            or self.text_index.key_count()
            or self.signature_index.key_count()
            or (self.chk_index is not None and self.chk_index.key_count())
        )

    def finish_content(self):
        """Finalize the pack content and compute the content hash name."""
        if self.name is not None:
            return
        self._writer.end()
        if self._buffer[1]:
            self._write_data(b"", flush=True)
        self.name = self._hash.hexdigest()

    def finish(self, suspend=False):
        """Finish the new pack.

        This:
         - finalises the content
         - assigns a name (the md5 of the content, currently)
         - writes out the associated indices
         - renames the pack into place.
         - stores the index size tuple for the pack in the index_sizes
           attribute.
        """
        self.finish_content()
        if not suspend:
            self._check_references()
        # write indices
        # XXX: It'd be better to write them all to temporary names, then
        # rename them all into place, so that the window when only some are
        # visible is smaller.  On the other hand none will be seen until
        # they're in the names list.
        self.index_sizes = [None, None, None, None]
        self._write_index("revision", self.revision_index, "revision", suspend)
        self._write_index("inventory", self.inventory_index, "inventory", suspend)
        self._write_index("text", self.text_index, "file texts", suspend)
        self._write_index(
            "signature", self.signature_index, "revision signatures", suspend
        )
        if self.chk_index is not None:
            self.index_sizes.append(None)
            self._write_index("chk", self.chk_index, "content hash bytes", suspend)
        self.write_stream.close(
            want_fdatasync=self._pack_collection.config_stack.get(
                "repository.fdatasync"
            )
        )
        # Note that this will clobber an existing pack with the same name,
        # without checking for hash collisions. While this is undesirable this
        # is something that can be rectified in a subsequent release. One way
        # to rectify it may be to leave the pack at the original name, writing
        # its pack-names entry as something like 'HASH: index-sizes
        # temporary-name'. Allocate that and check for collisions, if it is
        # collision free then rename it into place. If clients know this scheme
        # they can handle missing-file errors by:
        #  - try for HASH.pack
        #  - try for temporary-name
        #  - refresh the pack-list to see if the pack is now absent
        new_name = self.name + ".pack"
        if not suspend:
            new_name = "../packs/" + new_name
        self.upload_transport.move(self.random_name, new_name)
        self._state = "finished"
        # XXX: size might be interesting?
        logger.debug(
            "%s: create_pack: pack finished: %s%s->%s t+%6.3fs",
            time.ctime(),
            self.upload_transport.base,
            self.random_name,
            new_name,
            time.time() - self.start_time,
        )

    def flush(self):
        """Flush any current data."""
        if self._buffer[1]:
            bytes = b"".join(self._buffer[0])
            self.write_stream.write(bytes)
            self._hash.update(bytes)
            self._buffer[:] = [[], 0]

    def _get_external_refs(self, index):
        return index._external_references()

    def set_write_cache_size(self, size):
        """Set the write cache size in bytes."""
        self._cache_limit = size

    def _write_index(self, index_type, index, label, suspend=False):
        """Write out an index.

        :param index_type: The type of index to write - e.g. 'revision'.
        :param index: The index object to serialise.
        :param label: What label to give the index e.g. 'revision'.
        """
        index_name = self.index_name(index_type, self.name)
        transport = self.upload_transport if suspend else self.index_transport
        index_tempfile = index.finish()
        index_bytes = index_tempfile.read()
        write_stream = transport.open_write_stream(index_name, mode=self._file_mode)
        write_stream.write(index_bytes)
        write_stream.close(
            want_fdatasync=self._pack_collection.config_stack.get(
                "repository.fdatasync"
            )
        )
        self.index_sizes[self.index_offset(index_type)] = len(index_bytes)
        # XXX: size might be interesting?
        logger.debug(
            "%s: create_pack: wrote %s index: %s%s t+%6.3fs",
            time.ctime(),
            label,
            self.upload_transport.base,
            self.random_name,
            time.time() - self.start_time,
        )
        # Replace the writable index on this object with a readonly,
        # presently unloaded index. We should alter
        # the index layer to make its finish() error if add_node is
        # subsequently used. RBC
        self._replace_index_with_readonly(index_type)



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Inventory management for Bazaar.

This module provides classes and functions for managing file inventories
in Bazaar repositories, including entries for files, directories, symlinks,
and tree references.
"""

# FIXME: This refactoring of the workingtree code doesn't seem to keep
# the WorkingTree's copy of the inventory in sync with the branch.  The
# branch modifies its working inventory when it does a commit to make
# missing files permanently removed.

# TODO: Maybe also keep the full path of the entry, and the children?
# But those depend on its position within a particular inventory, and
# it would be nice not to need to hold the backpointer here.

__all__ = [
    "ROOT_ID",
    "CHKInventory",
    "FileId",
    "Inventory",
    "InventoryDirectory",
    "InventoryEntry",
    "InventoryFile",
    "InventoryLink",
    "TreeReference",
]

import contextlib
from collections import deque
from collections.abc import Iterable

from . import osutils
from ._bzr_rs import ROOT_ID
from ._bzr_rs import inventory as _mod_inventory_rs
from .errors import BadFileKindError, BzrFormatsError, InconsistentDelta


class NoSuchId(BzrFormatsError):
    """File ID not found in tree.

    Raised when a requested file ID is not present in the tree.
    """

    _fmt = 'The file id "%(file_id)s" is not present in the tree %(tree)s.'

    def __init__(self, tree, file_id):
        super().__init__()
        self.tree = tree
        self.file_id = file_id


FileId = bytes
InventoryEntry = _mod_inventory_rs.InventoryEntry
InventoryFile = _mod_inventory_rs.InventoryFile
InventoryDirectory = _mod_inventory_rs.InventoryDirectory
TreeReference = _mod_inventory_rs.TreeReference
InventoryLink = _mod_inventory_rs.InventoryLink
Inventory = _mod_inventory_rs.Inventory


class InvalidEntryName(BzrFormatsError):
    _fmt = "Invalid entry name: %(name)s"

    def __init__(self, name):
        super().__init__()
        self.name = name


class DuplicateFileId(BzrFormatsError):
    _fmt = "File id {%(file_id)s} already exists in inventory as %(entry)s"

    def __init__(self, file_id, entry):
        super().__init__()
        self.file_id = file_id
        self.entry = entry


class CHKInventory:
    """An inventory persisted in a CHK store.

    By design, a CHKInventory is immutable so many of the methods
    supported by Inventory - add, rename, apply_delta, etc - are *not*
    supported. To create a new CHKInventory, use create_by_apply_delta()
    or from_inventory(), say.

    Internally, a CHKInventory has one or two CHKMaps:

    * id_to_entry - a map from (file_id,) => InventoryEntry as bytes
    * parent_id_basename_to_file_id - a map from (parent_id, basename_utf8)
        => file_id as bytes

    The second map is optional and not present in early CHkRepository's.

    No caching is performed: every method call or item access will perform
    requests to the storage layer. As such, keep references to objects you
    want to reuse.
    """

    def has_filename(self, filename):
        """Check if a filename exists in the inventory.

        Args:
            filename: Path to check for existence.

        Returns:
            True if the filename exists, False otherwise.
        """
        return bool(self.path2id(filename))

    def id2path(self, file_id):
        """Return as a string the path to file_id.

        >>> i = Inventory()
        >>> e = i.add(InventoryDirectory(b'src-id', 'src', ROOT_ID))
        >>> e = i.add(InventoryFile(b'foo-id', 'foo.c', parent_id=b'src-id'))
        >>> print(i.id2path(b'foo-id'))
        src/foo.c

        :raises NoSuchId: If file_id is not present in the inventory.
        """
        # get all names, skipping root
        return "/".join(
            reversed(
                [parent.name for parent in self._iter_file_id_parents(file_id)][:-1]
            )
        )

    def iter_entries(self, from_dir=None, recursive=True):
        """Return (path, entry) pairs, in order by name.

        :param from_dir: if None, start from the root,
          otherwise start from this directory (either file-id or entry)
        :param recursive: recurse into directories or not
        """
        if from_dir is None:
            if self.root is None:
                return
            from_dir = self.root.file_id
            yield "", self.root
        elif not isinstance(from_dir, bytes):
            from_dir = from_dir.file_id

        # unrolling the recursive called changed the time from
        # 440ms/663ms (inline/total) to 116ms/116ms
        children = [(c.name, c) for c in self.iter_sorted_children(from_dir)]
        if not recursive:
            yield from children
            return
        children = deque(children)
        stack = [("", children)]
        while stack:
            from_dir_relpath, children = stack[-1]

            while children:
                name, ie = children.popleft()

                # we know that from_dir_relpath never ends in a slash
                # and 'f' doesn't begin with one, we can do a string op, rather
                # than the checks of pathjoin(), though this means that all paths
                # start with a slash
                path = from_dir_relpath + "/" + name

                yield path[1:], ie

                if ie.kind != "directory":
                    continue

                # But do this child first
                new_children = [
                    (c.name, c) for c in self.iter_sorted_children(ie.file_id)
                ]
                new_children = deque(new_children)
                stack.append((path, new_children))
                # Break out of inner loop, so that we start outer loop with child
                break
            else:
                # if we finished all children, pop it off the stack
                stack.pop()

    def iter_sorted_children(self, file_id):
        """Iterate through children of a directory in sorted order.

        Args:
            file_id: The file ID of the directory.

        Yields:
            Child inventory entries sorted by name.
        """
        return (c for (_n, c) in sorted(self.get_children(file_id).items()))

    def iter_entries_by_dir(self, from_dir=None, specific_file_ids=None):
        """Iterate over the entries in a directory first order.

        This returns all entries for a directory before returning
        the entries for children of a directory. This is not
        lexicographically sorted order, and is a hybrid between
        depth-first and breadth-first.

        :return: This yields (path, entry) pairs
        """
        if specific_file_ids and not isinstance(specific_file_ids, set):
            specific_file_ids = set(specific_file_ids)
        # TODO? Perhaps this should return the from_dir so that the root is
        # yielded? or maybe an option?
        if from_dir is None and specific_file_ids is None:
            # They are iterating from the root, and have not specified any
            # specific entries to look at. All current callers fully consume the
            # iterator, so we can safely assume we are accessing all entries
            self._preload_cache()
        if from_dir is None:
            if self.root is None:
                return
            # Optimize a common case
            if specific_file_ids is not None and len(specific_file_ids) == 1:
                file_id = list(specific_file_ids)[0]
                if file_id is not None:
                    try:
                        path = self.id2path(file_id)
                    except NoSuchId:
                        pass
                    else:
                        yield path, self.get_entry(file_id)
                return
            from_dir = self.root
            if specific_file_ids is None or self.root.file_id in specific_file_ids:
                yield "", self.root
        elif isinstance(from_dir, bytes):
            from_dir = self.get_entry(from_dir)
        else:
            raise TypeError(from_dir)

        if specific_file_ids is not None:
            # TODO: jam 20070302 This could really be done as a loop rather
            #       than a bunch of recursive calls.
            parents = set()
            byid = self

            def add_ancestors(file_id):
                if not byid.has_id(file_id):
                    return
                parent_id = byid.get_entry(file_id).parent_id
                if parent_id is None:
                    return
                if parent_id not in parents:
                    parents.add(parent_id)
                    add_ancestors(parent_id)

            for file_id in specific_file_ids:
                add_ancestors(file_id)
        else:
            parents = None

        stack = [("", from_dir)]
        while stack:
            cur_relpath, cur_dir = stack.pop()

            child_dirs = []
            for child_ie in self.iter_sorted_children(cur_dir.file_id):
                child_relpath = cur_relpath + child_ie.name

                if specific_file_ids is None or child_ie.file_id in specific_file_ids:
                    yield child_relpath, child_ie

                if child_ie.kind == "directory":
                    if parents is None or child_ie.file_id in parents:
                        child_dirs.append((child_relpath + "/", child_ie))
            stack.extend(reversed(child_dirs))

    def make_entry(self, kind, name, parent_id, file_id=None, revision=None, **kwargs):
        """Simple thunk to bzrformats.inventory.make_entry."""
        return make_entry(kind, name, parent_id, file_id, revision, **kwargs)

    def entries(self):
        """Return list of (path, ie) for all entries except the root.

        This may be faster than iter_entries.
        """
        accum = []

        def descend(dir_ie, dir_path):
            for ie in self.iter_sorted_children(dir_ie.file_id):
                child_path = osutils.pathjoin(dir_path, ie.name)
                accum.append((child_path, ie))
                if ie.kind == "directory":
                    descend(ie, child_path)

        if self.root is not None:
            descend(self.root, "")
        return accum

    def get_entry_by_path_partial(self, relpath):
        """Like get_entry_by_path, but return TreeReference objects.

        :param relpath: Path to resolve, either as string with / as separators,
            or as list of elements.
        :return: tuple with ie, resolved elements and elements left to resolve
        """
        names = osutils.splitpath(relpath) if isinstance(relpath, str) else relpath

        try:
            parent = self.root
        except NoSuchId:
            # root doesn't exist yet so nothing else can
            return None, None, None
        if parent is None:
            return None, None, None
        for i, f in enumerate(names):
            try:
                cie = self.get_child(parent.file_id, f)
                if cie is None:
                    return None, None, None
                if cie.kind == "tree-reference":
                    return cie, names[: i + 1], names[i + 1 :]
                parent = cie
            except KeyError:
                # or raise an error?
                return None, None, None
        return parent, names, []

    def get_entry_by_path(self, relpath):
        """Return an inventory entry by path.

        :param relpath: may be either a list of path components, or a single
            string, in which case it is automatically split.

        This returns the entry of the last component in the path,
        which may be either a file or a directory.

        Returns None IFF the path is not found.
        """
        names = osutils.splitpath(relpath) if isinstance(relpath, str) else relpath

        try:
            parent = self.root
        except NoSuchId:
            # root doesn't exist yet so nothing else can
            return None
        if parent is None:
            return None
        for f in names:
            try:
                cie = self.get_child(parent.file_id, f)
                if cie is None:
                    return None
                parent = cie
            except KeyError:
                # or raise an error?
                return None
        return parent

    def get_idpath(self, file_id):
        """Return a list of file_ids for the path to an entry.

        The list contains one element for each directory followed by
        the id of the file itself.  So the length of the returned list
        is equal to the depth of the file in the tree, counting the
        root directory as depth 1.
        """
        raise NotImplementedError(self.get_idpath)

    def __init__(self, search_key_name):
        """Initialize CHKInventory with a search key name.

        Args:
            search_key_name: Name of the search key for CHK operations.
        """
        self._fileid_to_entry_cache = {}
        self._fully_cached = False
        self._path_to_fileid_cache = {}
        self._search_key_name = search_key_name
        self.root_id = None
        self._children_cache = {}

    def __eq__(self, other):
        """Compare two sets by comparing their contents."""
        if not isinstance(other, CHKInventory):
            return NotImplemented

        this_key = self.id_to_entry.key()
        other_key = other.id_to_entry.key()
        this_pid_key = self.parent_id_basename_to_file_id.key()
        other_pid_key = other.parent_id_basename_to_file_id.key()
        if None in (this_key, this_pid_key, other_key, other_pid_key):
            return False
        return this_key == other_key and this_pid_key == other_pid_key

    def get_children(self, dir_id):
        """Access the list of children of this directory.

        With a parent_id_basename_to_file_id index, loads all the children,
        without loads the entire index. Without is bad. A more sophisticated
        proxy object might be nice, to allow partial loading of children as
        well when specific names are accessed. (So path traversal can be
        written in the obvious way but not examine siblings.).
        """
        children = self._children_cache.get(dir_id)
        if children is not None:
            return children
        # No longer supported
        if self.parent_id_basename_to_file_id is None:
            raise AssertionError(
                "Inventories without"
                " parent_id_basename_to_file_id are no longer supported"
            )
        result = {}
        # XXX: Todo - use proxy objects for the children rather than loading
        # all when the attribute is referenced.
        child_keys = set()
        for (
            _parent_id,
            _name_utf8,
        ), file_id in self.parent_id_basename_to_file_id.iteritems(
            key_filter=[(dir_id,)]
        ):
            child_keys.add((file_id,))
        cached = set()
        for file_id_key in child_keys:
            entry = self._fileid_to_entry_cache.get(file_id_key[0], None)
            if entry is not None:
                result[entry.name] = entry
                cached.add(file_id_key)
        child_keys.difference_update(cached)
        # populate; todo: do by name
        id_to_entry = self.id_to_entry
        for file_id_key, bytes in id_to_entry.iteritems(child_keys):
            entry = self._bytes_to_entry(bytes)
            result[entry.name] = entry
            self._fileid_to_entry_cache[file_id_key[0]] = entry
        self._children_cache[dir_id] = result
        return result

    def get_child(self, dir_id, name):
        """Get a specific child from a directory.

        Args:
            dir_id: The file ID of the directory.
            name: The name of the child to retrieve.

        Returns:
            The child inventory entry or None if not found.
        """
        # TODO(jelmer): Implement a version that doesn't load all children.
        return self.get_children(dir_id).get(name)

    def _expand_fileids_to_parents_and_children(self, file_ids):
        """Give a more wholistic view starting with the given file_ids.

        For any file_id which maps to a directory, we will include all children
        of that directory. We will also include all directories which are
        parents of the given file_ids, but we will not include their children.

        eg:
          /     # TREE_ROOT
          foo/  # foo-id
            baz # baz-id
            frob/ # frob-id
              fringle # fringle-id
          bar/  # bar-id
            bing # bing-id

        if given [foo-id] we will include
            TREE_ROOT as interesting parents
        and
            foo-id, baz-id, frob-id, fringle-id
        As interesting ids.
        """
        interesting = set()
        # TODO: Pre-pass over the list of fileids to see if anything is already
        #       deserialized in self._fileid_to_entry_cache

        directories_to_expand = set()
        children_of_parent_id = {}
        # It is okay if some of the fileids are missing
        for entry in self._getitems(file_ids):
            if entry.kind == "directory":
                directories_to_expand.add(entry.file_id)
            interesting.add(entry.parent_id)
            children_of_parent_id.setdefault(entry.parent_id, set()).add(entry.file_id)

        # Now, interesting has all of the direct parents, but not the
        # parents of those parents. It also may have some duplicates with
        # specific_fileids
        remaining_parents = interesting.difference(file_ids)
        # When we hit the TREE_ROOT, we'll get an interesting parent of None,
        # but we don't actually want to recurse into that
        interesting.add(None)  # this will auto-filter it in the loop
        remaining_parents.discard(None)
        while remaining_parents:
            next_parents = set()
            for entry in self._getitems(remaining_parents):
                next_parents.add(entry.parent_id)
                children_of_parent_id.setdefault(entry.parent_id, set()).add(
                    entry.file_id
                )
            # Remove any search tips we've already processed
            remaining_parents = next_parents.difference(interesting)
            interesting.update(remaining_parents)
            # We should probably also .difference(directories_to_expand)
        interesting.update(file_ids)
        interesting.discard(None)
        while directories_to_expand:
            # Expand directories by looking in the
            # parent_id_basename_to_file_id map
            keys = [(f,) for f in directories_to_expand]
            directories_to_expand = set()
            items = self.parent_id_basename_to_file_id.iteritems(keys)
            next_file_ids = {item[1] for item in items}
            next_file_ids = next_file_ids.difference(interesting)
            interesting.update(next_file_ids)
            for entry in self._getitems(next_file_ids):
                if entry.kind == "directory":
                    directories_to_expand.add(entry.file_id)
                children_of_parent_id.setdefault(entry.parent_id, set()).add(
                    entry.file_id
                )
        return interesting, children_of_parent_id

    def filter(self, specific_fileids):
        """Get an inventory view filtered against a set of file-ids.

        Children of directories and parents are included.

        The result may or may not reference the underlying inventory
        so it should be treated as immutable.
        """
        (
            interesting,
            parent_to_children,
        ) = self._expand_fileids_to_parents_and_children(specific_fileids)
        # There is some overlap here, but we assume that all interesting items
        # are in the _fileid_to_entry_cache because we had to read them to
        # determine if they were a dir we wanted to recurse, or just a file
        # This should give us all the entries we'll want to add, so start
        # adding
        other = Inventory(root_id=None)
        root = InventoryDirectory(self.root_id, "", None, self.root.revision)
        other.add(root)
        other.revision_id = self.revision_id
        if not interesting or not parent_to_children:
            # empty filter, or filtering entrys that don't exist
            # (if even 1 existed, then we would have populated
            # parent_to_children with at least the tree root.)
            return other
        cache = self._fileid_to_entry_cache
        remaining_children = deque(parent_to_children[self.root_id])
        while remaining_children:
            file_id = remaining_children.popleft()
            ie = cache[file_id]
            if ie.kind == "directory":
                ie = ie.copy()  # We create a copy to depopulate the .children attribute
            # TODO: depending on the uses of 'other' we should probably alwyas
            #       '.copy()' to prevent someone from mutating other and
            #       invaliding our internal cache
            other.add(ie)
            if file_id in parent_to_children:
                remaining_children.extend(parent_to_children[file_id])
        return other

    def _bytes_to_entry(self, bytes):
        """Deserialise a serialised entry."""
        result = _chk_inventory_bytes_to_entry(bytes)
        self._fileid_to_entry_cache[result.file_id] = result
        return result

    def create_by_apply_delta(
        self, inventory_delta, new_revision_id, propagate_caches=False
    ):
        """Create a new CHKInventory by applying inventory_delta to this one.

        See the inventory developers documentation for the theory behind
        inventory deltas.

        :param inventory_delta: The inventory delta to apply. See
            Inventory.apply_delta for details.
        :param new_revision_id: The revision id of the resulting CHKInventory.
        :param propagate_caches: If True, the caches for this inventory are
          copied to and updated for the result.
        :return: The new CHKInventory.
        """
        split = osutils.split
        result = CHKInventory(self._search_key_name)
        if propagate_caches:
            # Just propagate the path-to-fileid cache for now
            result._path_to_fileid_cache = self._path_to_fileid_cache.copy()
        from . import chk_map

        search_key_func = chk_map.search_key_registry.get(self._search_key_name)
        self.id_to_entry._ensure_root()
        maximum_size = self.id_to_entry._root_node.maximum_size
        result.revision_id = new_revision_id
        result.id_to_entry = chk_map.CHKMap(
            self.id_to_entry._store,
            self.id_to_entry.key(),
            search_key_func=search_key_func,
        )
        result.id_to_entry._ensure_root()
        result.id_to_entry._root_node.set_maximum_size(maximum_size)
        # Change to apply to the parent_id_basename delta. The dict maps
        # (parent_id, basename) -> (old_key, new_value). We use a dict because
        # when a path has its id replaced (e.g. the root is changed, or someone
        # does bzr mv a b, bzr mv c a, we should output a single change to this
        # map rather than two.
        parent_id_basename_delta = {}
        if self.parent_id_basename_to_file_id is not None:
            result.parent_id_basename_to_file_id = chk_map.CHKMap(
                self.parent_id_basename_to_file_id._store,
                self.parent_id_basename_to_file_id.key(),
                search_key_func=search_key_func,
            )
            result.parent_id_basename_to_file_id._ensure_root()
            self.parent_id_basename_to_file_id._ensure_root()
            result_p_id_root = result.parent_id_basename_to_file_id._root_node
            p_id_root = self.parent_id_basename_to_file_id._root_node
            result_p_id_root.set_maximum_size(p_id_root.maximum_size)
            result_p_id_root._key_width = p_id_root._key_width
        else:
            result.parent_id_basename_to_file_id = None
        result.root_id = self.root_id
        id_to_entry_delta = []
        # inventory_delta is only traversed once, so we just update the
        # variable.
        inventory_delta.check()
        # All changed entries need to have their parents be directories and be
        # at the right path. This set contains (path, id) tuples.
        parents = set()
        # When we delete an item, all the children of it must be either deleted
        # or altered in their own right. As we batch process the change via
        # CHKMap.apply_delta, we build a set of things to use to validate the
        # delta.
        deletes = set()
        altered = set()
        for old_path, new_path, file_id, entry in inventory_delta:
            # file id changes
            if new_path == "":
                result.root_id = file_id
            if new_path is None:
                # Make a delete:
                new_key = None
                new_value = None
                # Update caches
                if propagate_caches:
                    with contextlib.suppress(KeyError):
                        del result._path_to_fileid_cache[old_path]
                deletes.add(file_id)
            else:
                new_key = (file_id,)
                new_value = _chk_inventory_entry_to_bytes(entry)
                # Update caches. It's worth doing this whether
                # we're propagating the old caches or not.
                result._path_to_fileid_cache[new_path] = file_id
                parents.add((split(new_path)[0], entry.parent_id))
            if old_path is None:
                old_key = None
            else:
                old_key = (file_id,)
                if self.id2path(file_id) != old_path:
                    raise InconsistentDelta(
                        old_path,
                        file_id,
                        "Entry was at wrong other path {!r}.".format(
                            self.id2path(file_id)
                        ),
                    )
                altered.add(file_id)
            id_to_entry_delta.append((old_key, new_key, new_value))
            if result.parent_id_basename_to_file_id is not None:
                # parent_id, basename changes
                if old_path is None:
                    old_key = None
                else:
                    old_entry = self.get_entry(file_id)
                    old_key = self._parent_id_basename_key(old_entry)
                if new_path is None:
                    new_key = None
                    new_value = None
                else:
                    new_key = self._parent_id_basename_key(entry)
                    new_value = file_id
                # If the two keys are the same, the value will be unchanged
                # as its always the file id for this entry.
                if old_key != new_key:
                    # Transform a change into explicit delete/add preserving
                    # a possible match on the key from a different file id.
                    if old_key is not None:
                        parent_id_basename_delta.setdefault(old_key, [None, None])[
                            0
                        ] = old_key
                    if new_key is not None:
                        parent_id_basename_delta.setdefault(new_key, [None, None])[
                            1
                        ] = new_value
        # validate that deletes are complete.
        for file_id in deletes:
            entry = self.get_entry(file_id)
            if entry.kind != "directory":
                continue
            # This loop could potentially be better by using the id_basename
            # map to just get the child file ids.
            for child in self.iter_sorted_children(entry.file_id):
                if child.file_id not in altered:
                    raise InconsistentDelta(
                        self.id2path(child.file_id),
                        child.file_id,
                        "Child not deleted or reparented when parent deleted.",
                    )
        result.id_to_entry.apply_delta(id_to_entry_delta)
        if parent_id_basename_delta:
            # Transform the parent_id_basename delta data into a linear delta
            # with only one record for a given key. Optimally this would allow
            # re-keying, but its simpler to just output that as a delete+add
            # to spend less time calculating the delta.
            delta_list = []
            for key, (old_key, value) in parent_id_basename_delta.items():
                if value is not None:
                    delta_list.append((old_key, key, value))
                else:
                    delta_list.append((old_key, None, None))
            result.parent_id_basename_to_file_id.apply_delta(delta_list)
        parents.discard(("", None))
        for parent_path, parent in parents:
            try:
                if result.get_entry(parent).kind != "directory":
                    raise InconsistentDelta(
                        result.id2path(parent),
                        parent,
                        "Not a directory, but given children",
                    )
            except NoSuchId as e:
                raise InconsistentDelta(
                    "<unknown>", parent, "Parent is not present in resulting inventory."
                ) from e
            if result.path2id(parent_path) != parent:
                raise InconsistentDelta(
                    parent_path,
                    parent,
                    f"Parent has wrong path {result.path2id(parent_path)!r}.",
                )
        return result

    @classmethod
    def deserialise(klass, chk_store, lines, expected_revision_id):
        """Deserialise a CHKInventory.

        :param chk_store: A CHK capable VersionedFiles instance.
        :param bytes: The serialised bytes.
        :param expected_revision_id: The revision ID we think this inventory is
            for.
        :return: A CHKInventory
        """
        if not lines[-1].endswith(b"\n"):
            raise ValueError("last line should have trailing eol\n")
        if lines[0] != b"chkinventory:\n":
            raise ValueError(f"not a serialised CHKInventory: {bytes!r}")
        info = {}
        allowed_keys = frozenset(
            (
                b"root_id",
                b"revision_id",
                b"parent_id_basename_to_file_id",
                b"search_key_name",
                b"id_to_entry",
            )
        )
        for line in lines[1:]:
            key, value = line.rstrip(b"\n").split(b": ", 1)
            if key not in allowed_keys:
                raise BzrFormatsError(f"Unknown key in inventory: {key!r}\n{bytes!r}")
            if key in info:
                raise BzrFormatsError(f"Duplicate key in inventory: {key!r}\n{bytes!r}")
            info[key] = value
        revision_id = info[b"revision_id"]
        root_id = info[b"root_id"]
        search_key_name = info.get(b"search_key_name", b"plain")
        parent_id_basename_to_file_id = info.get(b"parent_id_basename_to_file_id")
        if not parent_id_basename_to_file_id.startswith(b"sha1:"):
            raise ValueError(
                "parent_id_basename_to_file_id should be a sha1"
                f" key not {parent_id_basename_to_file_id!r}"
            )
        id_to_entry = info[b"id_to_entry"]
        if not id_to_entry.startswith(b"sha1:"):
            raise ValueError(f"id_to_entry should be a sha1 key not {id_to_entry!r}")

        result = CHKInventory(search_key_name)
        result.revision_id = revision_id
        result.root_id = root_id
        from . import chk_map

        search_key_func = chk_map.search_key_registry.get(result._search_key_name)
        if parent_id_basename_to_file_id is not None:
            result.parent_id_basename_to_file_id = chk_map.CHKMap(
                chk_store,
                (parent_id_basename_to_file_id,),
                search_key_func=search_key_func,
            )
        else:
            result.parent_id_basename_to_file_id = None

        result.id_to_entry = chk_map.CHKMap(
            chk_store,
            (id_to_entry,),
            search_key_func=search_key_func,
        )
        if (result.revision_id,) != expected_revision_id:
            raise ValueError(
                f"Mismatched revision id and expected: {result.revision_id!r}, {expected_revision_id!r}"
            )
        return result

    @classmethod
    def from_inventory(
        klass, chk_store, inventory, maximum_size=0, search_key_name=b"plain"
    ):
        """Create a CHKInventory from an existing inventory.

        The content of inventory is copied into the chk_store, and a
        CHKInventory referencing that is returned.

        :param chk_store: A CHK capable VersionedFiles instance.
        :param inventory: The inventory to copy.
        :param maximum_size: The CHKMap node size limit.
        :param search_key_name: The identifier for the search key function
        """
        result = klass(search_key_name)
        result.revision_id = inventory.revision_id
        result.root_id = inventory.root.file_id

        parent_id_basename_key = result._parent_id_basename_key
        id_to_entry_dict = {}
        parent_id_basename_dict = {}
        for _path, entry in inventory.iter_entries():
            key = (entry.file_id,)
            id_to_entry_dict[key] = _chk_inventory_entry_to_bytes(entry)
            p_id_key = parent_id_basename_key(entry)
            parent_id_basename_dict[p_id_key] = entry.file_id

        result._populate_from_dicts(
            chk_store,
            id_to_entry_dict,
            parent_id_basename_dict,
            maximum_size=maximum_size,
        )
        return result

    def _populate_from_dicts(
        self, chk_store, id_to_entry_dict, parent_id_basename_dict, maximum_size
    ):
        from . import chk_map

        search_key_func = chk_map.search_key_registry.get(self._search_key_name)
        root_key = chk_map.CHKMap.from_dict(
            chk_store,
            id_to_entry_dict,
            maximum_size=maximum_size,
            key_width=1,
            search_key_func=search_key_func,
        )
        self.id_to_entry = chk_map.CHKMap(chk_store, root_key, search_key_func)
        root_key = chk_map.CHKMap.from_dict(
            chk_store,
            parent_id_basename_dict,
            maximum_size=maximum_size,
            key_width=2,
            search_key_func=search_key_func,
        )
        self.parent_id_basename_to_file_id = chk_map.CHKMap(
            chk_store, root_key, search_key_func
        )

    def _parent_id_basename_key(self, entry):
        """Create a key for a entry in a parent_id_basename_to_file_id index."""
        parent_id = entry.parent_id if entry.parent_id is not None else b""
        return (parent_id, entry.name.encode("utf8"))

    def get_entry(self, file_id):
        """Map a single file_id -> InventoryEntry."""
        if file_id is None:
            raise NoSuchId(self, file_id)
        result = self._fileid_to_entry_cache.get(file_id, None)
        if result is not None:
            return result
        try:
            return self._bytes_to_entry(
                next(self.id_to_entry.iteritems([(file_id,)]))[1]
            )
        except StopIteration as e:
            # really we're passing an inventory, not a tree...
            raise NoSuchId(self, file_id) from e

    def _getitems(self, file_ids: Iterable[FileId]) -> list[InventoryEntry]:  # type: ignore
        """Similar to get_entry, but lets you query for multiple.

        The returned order is undefined. And currently if an item doesn't
        exist, it isn't included in the output.
        """
        result: list[InventoryEntry] = []  # type: ignore
        remaining: list[FileId] = []
        for file_id in file_ids:
            entry = self._fileid_to_entry_cache.get(file_id, None)
            if entry is None:
                remaining.append(file_id)
            else:
                result.append(entry)
        file_keys = [(f,) for f in remaining]
        for _file_key, value in self.id_to_entry.iteritems(file_keys):
            entry = self._bytes_to_entry(value)
            result.append(entry)
            self._fileid_to_entry_cache[entry.file_id] = entry
        return result

    def has_id(self, file_id):
        """Check if a file_id exists in the inventory.

        Args:
            file_id: The file ID to check.

        Returns:
            True if the file_id exists, False otherwise.
        """
        # Perhaps have an explicit 'contains' method on CHKMap ?
        if self._fileid_to_entry_cache.get(file_id, None) is not None:
            return True
        return len(list(self.id_to_entry.iteritems([(file_id,)]))) == 1

    def is_root(self, file_id):
        """Check if a file_id is the root of the inventory.

        Args:
            file_id: The file ID to check.

        Returns:
            True if this is the root file ID, False otherwise.
        """
        return file_id == self.root_id

    def _iter_file_id_parents(self, file_id):
        """Yield the parents of file_id up to the root."""
        while file_id is not None:
            try:
                ie = self.get_entry(file_id)
            except KeyError as e:
                raise NoSuchId(tree=self, file_id=file_id) from e
            yield ie
            file_id = ie.parent_id

    def iter_all_ids(self):
        """Iterate over all file-ids."""
        for key, _ in self.id_to_entry.iteritems():
            yield key[-1]

    def iter_just_entries(self):
        """Iterate over all entries.

        Unlike iter_entries(), just the entries are returned (not (path, ie))
        and the order of entries is undefined.

        XXX: We may not want to merge this into bzr.dev.
        """
        for key, entry in self.id_to_entry.iteritems():
            file_id = key[0]
            ie = self._fileid_to_entry_cache.get(file_id, None)
            if ie is None:
                ie = self._bytes_to_entry(entry)
                self._fileid_to_entry_cache[file_id] = ie
            yield ie

    def _preload_cache(self):
        """Make sure all file-ids are in _fileid_to_entry_cache."""
        if self._fully_cached:
            return  # No need to do it again
        # The optimal sort order is to use iteritems() directly
        cache = self._fileid_to_entry_cache
        for key, entry in self.id_to_entry.iteritems():
            file_id = key[0]
            if file_id not in cache:
                ie = self._bytes_to_entry(entry)
                cache[file_id] = ie
            else:
                ie = cache[file_id]
        last_parent_id = last_parent_ie = None
        pid_items = self.parent_id_basename_to_file_id.iteritems()
        for key, child_file_id in pid_items:
            if key == (b"", b""):  # This is the root
                if child_file_id != self.root_id:
                    raise ValueError(
                        "Data inconsistency detected."
                        ' We expected data with key ("","") to match'
                        f" the root id, but {child_file_id} != {self.root_id}"
                    )
                continue
            parent_id, basename = key
            ie = cache[child_file_id]
            if parent_id == last_parent_id:
                if last_parent_ie is None:
                    raise AssertionError("last_parent_ie should not be None")
                parent_ie = last_parent_ie
            else:
                parent_ie = cache[parent_id]
            if parent_ie.kind != "directory":
                raise ValueError(
                    "Data inconsistency detected."
                    " An entry in the parent_id_basename_to_file_id map"
                    f" has parent_id {{{parent_id}}} but the kind of that object"
                    f' is {parent_ie.kind!r} not "directory"'
                )
            siblings = self._children_cache.setdefault(parent_ie.file_id, {})
            basename = basename.decode("utf-8")
            if basename in siblings:
                existing_ie = siblings[basename]
                if existing_ie != ie:
                    raise ValueError(
                        "Data inconsistency detected."
                        f" Two entries with basename {basename!r} were found"
                        f" in the parent entry {{{parent_id}}}"
                    )
            if basename != ie.name:
                raise ValueError(
                    "Data inconsistency detected."
                    " In the parent_id_basename_to_file_id map, file_id"
                    " {{{}}} is listed as having basename {!r}, but in the"
                    " id_to_entry map it is {!r}".format(
                        child_file_id, basename, ie.name
                    )
                )
            siblings[basename] = ie
        self._fully_cached = True

    def iter_changes(self, basis):
        """Generate a Tree.iter_changes change list between this and basis.

        :param basis: Another CHKInventory.
        :return: An iterator over the changes between self and basis, as per
            tree.iter_changes().
        """
        # We want: (file_id, (path_in_source, path_in_target),
        # changed_content, versioned, parent, name, kind,
        # executable)
        for key, basis_value, self_value in self.id_to_entry.iter_changes(
            basis.id_to_entry
        ):
            file_id = key[0]
            if basis_value is not None:
                basis_entry = basis._bytes_to_entry(basis_value)
                path_in_source = basis.id2path(file_id)
                basis_parent = basis_entry.parent_id
                basis_name = basis_entry.name
                basis_executable = basis_entry.executable
            else:
                path_in_source = None
                basis_parent = None
                basis_name = None
                basis_executable = None
            if self_value is not None:
                self_entry = self._bytes_to_entry(self_value)
                path_in_target = self.id2path(file_id)
                self_parent = self_entry.parent_id
                self_name = self_entry.name
                self_executable = self_entry.executable
            else:
                path_in_target = None
                self_parent = None
                self_name = None
                self_executable = None
            if basis_value is None:
                # add
                kind = (None, self_entry.kind)
                versioned = (False, True)
            elif self_value is None:
                # delete
                kind = (basis_entry.kind, None)
                versioned = (True, False)
            else:
                kind = (basis_entry.kind, self_entry.kind)
                versioned = (True, True)
            changed_content = False
            if kind[0] != kind[1]:
                changed_content = True
            elif kind[0] == "file":
                if (
                    self_entry.text_size != basis_entry.text_size
                    or self_entry.text_sha1 != basis_entry.text_sha1
                ):
                    changed_content = True
            elif kind[0] == "symlink":
                if self_entry.symlink_target != basis_entry.symlink_target:
                    changed_content = True
            elif kind[0] == "tree-reference":
                if self_entry.reference_revision != basis_entry.reference_revision:
                    changed_content = True
            parent = (basis_parent, self_parent)
            name = (basis_name, self_name)
            executable = (basis_executable, self_executable)
            if (
                not changed_content
                and parent[0] == parent[1]
                and name[0] == name[1]
                and executable[0] == executable[1]
            ):
                # Could happen when only the revision changed for a directory
                # for instance.
                continue
            yield (
                file_id,
                (path_in_source, path_in_target),
                changed_content,
                versioned,
                parent,
                name,
                kind,
                executable,
            )

    def __len__(self) -> int:
        """Return the number of entries in the inventory."""
        return len(self.id_to_entry)

    def path2id(self, relpath):
        """Return the file ID for a relative path.

        Args:
            relpath: Relative path as string or list of path components.

        Returns:
            The file ID for the path, or None if not found.
        """
        # TODO: perhaps support negative hits?
        if isinstance(relpath, str):
            names = osutils.splitpath(relpath)
        else:
            names = relpath
            if relpath == []:
                relpath = [""]
            relpath = osutils.pathjoin(*relpath)
        result = self._path_to_fileid_cache.get(relpath, None)
        if result is not None:
            return result
        current_id = self.root_id
        if current_id is None:
            return None
        parent_id_index = self.parent_id_basename_to_file_id
        cur_path = None
        for basename in names:
            cur_path = basename if cur_path is None else cur_path + "/" + basename
            basename_utf8 = basename.encode("utf8")
            file_id = self._path_to_fileid_cache.get(cur_path, None)
            if file_id is None:
                key_filter = [(current_id, basename_utf8)]
                items = parent_id_index.iteritems(key_filter)
                for (parent_id, name_utf8), file_id in items:  # noqa: B007
                    if parent_id != current_id or name_utf8 != basename_utf8:
                        raise BzrFormatsError(
                            "corrupt inventory lookup! {!r} {!r} {!r} {!r}".format(
                                parent_id, current_id, name_utf8, basename_utf8
                            )
                        )
                if file_id is None:
                    return None
                else:
                    self._path_to_fileid_cache[cur_path] = file_id
            current_id = file_id
        return current_id

    def to_lines(self):
        """Serialise the inventory to lines."""
        lines = [b"chkinventory:\n"]
        if self._search_key_name != b"plain":
            # custom ordering grouping things that don't change together
            lines.append(b"search_key_name: %s\n" % (self._search_key_name))
            lines.append(b"root_id: %s\n" % self.root_id)
            lines.append(
                b"parent_id_basename_to_file_id: %s\n"
                % (self.parent_id_basename_to_file_id.key()[0],)
            )
            lines.append(b"revision_id: %s\n" % self.revision_id)
            lines.append(b"id_to_entry: %s\n" % (self.id_to_entry.key()[0],))
        else:
            lines.append(b"revision_id: %s\n" % self.revision_id)
            lines.append(b"root_id: %s\n" % self.root_id)
            if self.parent_id_basename_to_file_id is not None:
                lines.append(
                    b"parent_id_basename_to_file_id: %s\n"
                    % (self.parent_id_basename_to_file_id.key()[0],)
                )
            lines.append(b"id_to_entry: %s\n" % (self.id_to_entry.key()[0],))
        return lines

    @property
    def root(self):
        """Get the root entry."""
        return self.get_entry(self.root_id)


entry_factory = {
    "directory": InventoryDirectory,
    "file": InventoryFile,
    "symlink": InventoryLink,
    "tree-reference": TreeReference,
}


def make_entry(kind, name, parent_id, file_id=None, revision=None, **kwargs):
    """Create an inventory entry.

    :param kind: the type of inventory entry to create.
    :param name: the basename of the entry.
    :param parent_id: the parent_id of the entry.
    :param file_id: the file_id to use. if None, one will be created.
    """
    if file_id is None:
        from . import generate_ids

        file_id = generate_ids.gen_file_id(name)
    name = ensure_normalized_name(name)
    try:
        factory = entry_factory[kind]
    except KeyError as e:
        raise BadFileKindError(name, kind) from e
    return factory(file_id, name, parent_id, revision, **kwargs)


ensure_normalized_name = _mod_inventory_rs.ensure_normalized_name
is_valid_name = _mod_inventory_rs.is_valid_name


def mutable_inventory_from_tree(tree):
    """Create a new inventory that has the same contents as a specified tree.

    :param tree: Revision tree to create inventory from
    """
    entries = tree.iter_entries_by_dir()
    inv = Inventory(None, tree.get_revision_id())
    for _path, inv_entry in entries:
        inv.add(inv_entry.copy())
    return inv


chk_inventory_bytes_to_utf8name_key = (
    _mod_inventory_rs.chk_inventory_bytes_to_utf8name_key
)
_chk_inventory_bytes_to_entry = _mod_inventory_rs.chk_inventory_bytes_to_entry
_chk_inventory_entry_to_bytes = _mod_inventory_rs.chk_inventory_entry_to_bytes


def _make_delta(new, old):
    """Make an inventory delta from two inventories."""
    from .inventory_delta import InventoryDelta

    if isinstance(old, CHKInventory) and isinstance(new, CHKInventory):
        delta = []
        for key, old_value, self_value in new.id_to_entry.iter_changes(old.id_to_entry):
            file_id = key[0]
            old_path = old.id2path(file_id) if old_value is not None else None
            if self_value is not None:
                entry = new._bytes_to_entry(self_value)
                new._fileid_to_entry_cache[file_id] = entry
                new_path = new.id2path(file_id)
            else:
                entry = None
                new_path = None
            delta.append((old_path, new_path, file_id, entry))
        return InventoryDelta(delta)
    elif isinstance(old, Inventory) and isinstance(new, Inventory):
        return new._make_delta(old)
    else:
        old_ids = set(old.iter_all_ids())
        new_ids = set(new.iter_all_ids())
        adds = new_ids - old_ids
        deletes = old_ids - new_ids
        common = old_ids.intersection(new_ids)
        delta = []
        for file_id in deletes:
            delta.append((old.id2path(file_id), None, file_id, None))
        for file_id in adds:
            delta.append((None, new.id2path(file_id), file_id, new.get_entry(file_id)))
        for file_id in common:
            if old.get_entry(file_id) != new.get_entry(file_id):
                delta.append(
                    (
                        old.id2path(file_id),
                        new.id2path(file_id),
                        file_id,
                        new.get_entry(file_id),
                    )
                )
        return InventoryDelta(delta)



# Copyright (C) 2025 Breezy Contributors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""OS utilities for bzrformats using only standard library."""

import hashlib
import logging
import os
import shutil
import sys
import unicodedata


def split(path):
    """Split a pathname into directory and basename parts."""
    if isinstance(path, bytes):
        return os.path.split(path)
    else:
        # For unicode strings, encode to UTF-8, split, then decode
        encoded = path.encode("utf-8")
        dirname, basename = os.path.split(encoded)
        return dirname.decode("utf-8"), basename.decode("utf-8")


def pathjoin(*args):
    """Join paths together."""
    if not args:
        return b"" if isinstance(args[0], bytes) else ""

    # Check if we're dealing with bytes or strings
    if isinstance(args[0], bytes):
        return os.path.join(*args)
    else:
        # For unicode strings, encode to UTF-8, join, then decode
        encoded_args = [arg.encode("utf-8") for arg in args]
        result = os.path.join(*encoded_args)
        return result.decode("utf-8")


def pumpfile(from_file, to_file, buffer_size=65536):
    """Copy data from one file-like object to another.

    Returns the number of bytes copied.
    """
    initial_pos = to_file.tell() if hasattr(to_file, "tell") else 0
    shutil.copyfileobj(from_file, to_file, buffer_size)
    if hasattr(to_file, "tell"):
        return to_file.tell() - initial_pos
    else:
        # If we can't tell the position, we can't return accurate byte count
        return 0


def chunks_to_lines(chunks):
    """Convert chunks to lines."""
    if not chunks:
        return []

    # Join all chunks
    data = b"".join(chunks)

    # Split into lines, keeping line endings
    lines = []
    start = 0
    for i, byte in enumerate(data):
        if byte == ord(b"\n"):
            lines.append(data[start : i + 1])
            start = i + 1

    # Add remaining data if any
    if start < len(data):
        lines.append(data[start:])

    return lines


def normalized_filename(filename):
    """Return the normalized form of a filename.

    Returns (normalized_name, can_access) tuple.
    """
    if isinstance(filename, bytes):
        # For bytes, try to decode as UTF-8 first
        try:
            unicode_filename = filename.decode("utf-8")
        except UnicodeDecodeError:
            # If it's not valid UTF-8, return as-is
            return filename, True
    else:
        unicode_filename = filename

    # Normalize using NFC (Canonical Decomposition, followed by Canonical Composition)
    normalized = unicodedata.normalize("NFC", unicode_filename)

    if isinstance(filename, bytes):
        try:
            return normalized.encode("utf-8"), True
        except UnicodeEncodeError:
            return filename, True
    else:
        return normalized, True


def failed_to_load_extension(exception):
    """Log a message about a failed extension load."""
    logging.debug("Failed to load extension: %s", exception)


def fdatasync(fileno):
    """Flush file contents to disk, not metadata."""
    try:
        os.fdatasync(fileno)
    except AttributeError:
        # fdatasync is not available on all platforms (e.g., Windows)
        # Fall back to fsync which is more widely available
        os.fsync(fileno)


def splitpath(path):
    """Split a path into a list of components."""
    if isinstance(path, bytes):
        if path.startswith(b"/"):
            path = path[1:]
        if not path:
            return []
        return path.split(b"/")
    else:
        if path.startswith("/"):
            path = path[1:]
        if not path:
            return []
        return path.split("/")


def file_kind_from_stat_mode(mode):
    """Return the file kind based on the stat mode."""
    import stat

    if stat.S_ISREG(mode):
        return "file"
    elif stat.S_ISDIR(mode):
        return "directory"
    elif stat.S_ISLNK(mode):
        return "symlink"
    elif stat.S_ISFIFO(mode):
        return "fifo"
    elif stat.S_ISSOCK(mode):
        return "socket"
    elif stat.S_ISCHR(mode):
        return "chardev"
    elif stat.S_ISBLK(mode):
        return "block"
    else:
        return "unknown"


def contains_whitespace(s):
    """Return True if the string contains whitespace characters."""
    # Check for common whitespace characters
    if isinstance(s, bytes):
        return any(c in s for c in b" \t\n\r\v\f")
    else:
        return any(c in s for c in " \t\n\r\v\f")


def sha_strings(strings):
    """Return the sha1 of concatenated strings."""
    sha = hashlib.sha1()  # noqa: S324
    for string in strings:
        if isinstance(string, str):
            # Convert unicode strings to bytes using UTF-8
            string = string.encode("utf-8")
        sha.update(string)
    return sha.hexdigest().encode("ascii")


def sha_string(string):
    """Return the sha1 of a single string."""
    if isinstance(string, str):
        # Convert unicode strings to bytes using UTF-8
        string = string.encode("utf-8")
    sha = hashlib.sha1()  # noqa: S324
    sha.update(string)
    return sha.hexdigest().encode("ascii")


def sha_file(file_obj):
    """Return the sha1 of a file."""
    sha = hashlib.sha1()  # noqa: S324
    while True:
        chunk = file_obj.read(65536)
        if not chunk:
            break
        sha.update(chunk)
    return sha.hexdigest().encode("ascii")


def dirname(path):
    """Return the directory part of a path."""
    if isinstance(path, bytes):
        return os.path.dirname(path)
    else:
        # For unicode strings, encode to UTF-8, get dirname, then decode
        encoded = path.encode("utf-8")
        result = os.path.dirname(encoded)
        return result.decode("utf-8")


def basename(path):
    """Return the basename part of a path."""
    if isinstance(path, bytes):
        return os.path.basename(path)
    else:
        # For unicode strings, encode to UTF-8, get basename, then decode
        encoded = path.encode("utf-8")
        result = os.path.basename(encoded)
        return result.decode("utf-8")


def chunks_to_lines_iter(chunks_iter):
    """Convert an iterator of chunks to an iterator of lines."""
    buffer = b""
    for chunk in chunks_iter:
        buffer += chunk
        while b"\n" in buffer:
            line, buffer = buffer.split(b"\n", 1)
            yield line + b"\n"

    # Yield any remaining data as the last line (without newline)
    if buffer:
        yield buffer


def file_iterator(file_obj, chunk_size=65536):
    """Iterate over the contents of a file in chunks."""
    while True:
        chunk = file_obj.read(chunk_size)
        if not chunk:
            break
        yield chunk


def normalizes_filenames():
    """Check if the filesystem normalizes filenames (e.g. Mac OS X)."""
    from . import _osutils_rs

    return _osutils_rs.normalizes_filenames()


def rand_chars(length):
    """Generate a string of random characters."""
    from . import _osutils_rs

    return _osutils_rs.rand_chars(length)


class DirReader:
    """An interface for reading directories."""

    def top_prefix_to_starting_dir(self, top, prefix=""):
        """Converts top and prefix to a starting dir entry.

        :param top: A utf8 path
        :param prefix: An optional utf8 path to prefix output relative paths
            with.
        :return: A tuple starting with prefix, and ending with the native
            encoding of top.
        """
        raise NotImplementedError(self.top_prefix_to_starting_dir)

    def read_dir(self, prefix, top):
        """Read a specific dir.

        :param prefix: A utf8 prefix to be preprended to the path basenames.
        :param top: A natively encoded path to read.
        :return: A list of the directories contents. Each item contains:
            (utf8_relpath, utf8_name, kind, lstatvalue, native_abspath)
        """
        raise NotImplementedError(self.read_dir)


_selected_dir_reader = None


def safe_unicode(unicode_or_utf8_string):
    """Coerce unicode_or_utf8_string into unicode.

    If it is unicode, it is returned.
    Otherwise it is decoded from utf-8. If decoding fails, the exception is
    wrapped in a TypeError exception.
    """
    if isinstance(unicode_or_utf8_string, (str, os.PathLike)):
        return unicode_or_utf8_string
    try:
        return unicode_or_utf8_string.decode("utf8")
    except UnicodeDecodeError as e:
        raise TypeError(unicode_or_utf8_string) from e


def safe_utf8(unicode_or_utf8_string):
    """Coerce unicode_or_utf8_string to a utf8 string.

    If it is a str, it is returned.
    If it is Unicode, it is encoded into a utf-8 string.
    """
    if isinstance(unicode_or_utf8_string, bytes):
        # Make sure it is a valid utf-8 string
        try:
            unicode_or_utf8_string.decode("utf-8")
        except UnicodeDecodeError as e:
            raise TypeError(unicode_or_utf8_string) from e
        return unicode_or_utf8_string
    return unicode_or_utf8_string.encode("utf-8")


def _walkdirs_utf8(top, prefix="", fs_enc=None):
    """Yield data about all the directories in a tree.

    This yields the same information as walkdirs() only each entry is yielded
    in utf-8. On platforms which have a filesystem encoding of utf8 the paths
    are returned as exact byte-strings.

    :return: yields a tuple of (dir_info, [file_info])
        dir_info is (utf8_relpath, path-from-top)
        file_info is (utf8_relpath, utf8_name, kind, lstat, path-from-top)
        if top is an absolute path, path-from-top is also an absolute path.
        path-from-top might be unicode or utf8, but it is the correct path to
        pass to os functions to affect the file in question. (such as os.lstat)
    """
    global _selected_dir_reader
    if _selected_dir_reader is None:
        if fs_enc is None:
            fs_enc = sys.getfilesystemencoding()
        # Always use the python version for bzrformats
        _selected_dir_reader = UnicodeDirReader()

    # 0 - relpath, 1- basename, 2- kind, 3- stat, 4-toppath
    # But we don't actually uses 1-3 in pending, so set them to None
    pending = [[_selected_dir_reader.top_prefix_to_starting_dir(top, prefix)]]
    read_dir = _selected_dir_reader.read_dir
    _directory = "directory"
    while pending:
        relroot, _, _, _, top = pending[-1].pop()
        if not pending[-1]:
            pending.pop()
        dirblock = sorted(read_dir(relroot, top))
        yield (relroot, top), dirblock
        # push the user specified dirs from dirblock
        next = [d for d in reversed(dirblock) if d[2] == _directory]
        if next:
            pending.append(next)


class UnicodeDirReader(DirReader):
    """A dir reader for non-utf8 file systems, which transcodes."""

    __slots__ = ["_utf8_encode"]

    def __init__(self):
        """Initialize the UTF-8 directory reader."""
        import codecs

        self._utf8_encode = codecs.getencoder("utf8")

    def top_prefix_to_starting_dir(self, top, prefix=""):
        """See DirReader.top_prefix_to_starting_dir."""
        return (safe_utf8(prefix), None, None, None, safe_unicode(top))

    def read_dir(self, prefix, top):
        """Read a single directory from a non-utf8 file system.

        top, and the abspath element in the output are unicode, all other paths
        are utf8. Local disk IO is done via unicode calls to listdir etc.

        This is currently the fallback code path when the filesystem encoding is
        not UTF-8. It may be better to implement an alternative so that we can
        safely handle paths that are not properly decodable in the current
        encoding.

        See DirReader.read_dir for details.
        """
        _utf8_encode = self._utf8_encode

        relprefix = prefix + b"/" if prefix else b""
        top_slash = top + "/"

        dirblock = []
        append = dirblock.append
        for entry in os.scandir(top):
            name = os.fsdecode(entry.name)
            abspath = top_slash + name
            name_utf8 = _utf8_encode(name, "surrogateescape")[0]
            statvalue = entry.stat(follow_symlinks=False)
            kind = file_kind_from_stat_mode(statvalue.st_mode)
            append((relprefix + name_utf8, name_utf8, kind, statvalue, abspath))
        return sorted(dirblock)


def is_inside(dir, fname):
    """Check if fname is inside dir.

    The empty string as dir is considered to contain everything.
    A path is considered to be inside itself.

    :param dir: Directory path (bytes or str)
    :param fname: File path to check (bytes or str)
    :return: True if fname is inside dir
    """
    # Normalize to use bytes for comparison
    if isinstance(dir, str):
        dir = dir.encode("utf-8")
    if isinstance(fname, str):
        fname = fname.encode("utf-8")

    if dir == fname:
        return True

    # Ensure trailing slash for proper comparison
    if dir != b"":
        dir = dir.rstrip(b"/") + b"/"

    return fname.startswith(dir)


def is_inside_any(dir_list, fname):
    """Check if fname is inside any of the directories in dir_list.

    :param dir_list: List of directory paths
    :param fname: File path to check
    :return: True if fname is inside any directory in dir_list
    """
    return any(is_inside(dir, fname) for dir in dir_list)


def parent_directories(filename):
    """Return a list of parent directories of filename.

    :param filename: Path (bytes or str)
    :return: List of parent directory paths
    """
    from . import _osutils_rs

    if isinstance(filename, bytes):
        filename = filename.decode("utf-8")
    return _osutils_rs.parent_directories(filename)


def split_lines(text):
    r"""Split text into lines, keeping line endings.

    Args:
        text: bytes to split

    Returns:
        List of byte strings, each ending with \\n where appropriate
    """
    from . import _osutils_rs

    return _osutils_rs.split_lines(text)


class IterableFile:
    """A file-like object backed by an iterator of byte strings.

    Supports ``read()`` and ``readline()`` over a lazy sequence of chunks.
    """

    def __init__(self, iterable):
        """Initialize with an iterable of byte chunks."""
        self._iter = iter(iterable)
        self._buf = b""

    def read(self, size=-1):
        """Read up to *size* bytes, or all remaining if *size* < 0."""
        if size < 0:
            return self._buf + b"".join(self._iter)
        while len(self._buf) < size:
            try:
                self._buf += next(self._iter)
            except StopIteration:
                break
        result = self._buf[:size]
        self._buf = self._buf[size:]
        return result

    def readline(self):
        r"""Read one line (up to and including ``\\n``)."""
        while b"\n" not in self._buf:
            try:
                self._buf += next(self._iter)
            except StopIteration:
                # Return whatever is left
                result = self._buf
                self._buf = b""
                return result
        idx = self._buf.index(b"\n") + 1
        result = self._buf[:idx]
        self._buf = self._buf[idx:]
        return result

    def readlines(self):
        """Return all remaining lines as a list."""
        lines = []
        while True:
            line = self.readline()
            if not line:
                break
            lines.append(line)
        return lines

    def __iter__(self):
        """Iterate over lines."""
        while True:
            line = self.readline()
            if not line:
                break
            yield line



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Inventory/revision serialization."""

from . import registry
from .errors import BzrFormatsError


class BadInventoryFormat(BzrFormatsError):
    """Base exception class for inventory serialization errors."""

    _fmt = "Root class for inventory serialization errors"


class UnexpectedInventoryFormat(BadInventoryFormat):
    """Raised when an inventory is not in the expected format."""

    _fmt = "The inventory was not in the expected format:\n %(msg)s"

    def __init__(self, msg):
        """Initialize UnexpectedInventoryFormat exception.

        Args:
            msg: Error message describing the unexpected format.
        """
        super().__init__(msg=msg)


class UnsupportedInventoryKind(BzrFormatsError):
    """Raised when an unsupported inventory entry kind is encountered."""

    _fmt = """Unsupported entry kind %(kind)s"""

    def __init__(self, kind):
        """Initialize UnsupportedInventoryKind exception.

        Args:
            kind: The unsupported entry kind.
        """
        super().__init__()
        self.kind = kind


class RevisionSerializer:
    """Revision serialization/deserialization."""

    squashes_xml_invalid_characters = False

    def write_revision_to_string(self, rev):
        """Serialize a revision to a string.

        Args:
            rev: The revision object to serialize.

        Returns:
            Serialized revision as a string.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError(self.write_revision_to_string)

    def write_revision_to_lines(self, rev):
        """Serialize a revision to a list of lines.

        Args:
            rev: The revision object to serialize.

        Returns:
            Serialized revision as a list of lines.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError(self.write_revision_to_lines)

    def read_revision(self, f):
        """Read a revision from a file object.

        Args:
            f: File-like object to read from.

        Returns:
            Deserialized revision object.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError(self.read_revision)

    def read_revision_from_string(self, xml_string):
        """Read a revision from a string.

        Args:
            xml_string: String containing the serialized revision.

        Returns:
            Deserialized revision object.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError(self.read_revision_from_string)


class InventorySerializer:
    """Inventory serialization/deserialization."""

    def write_inventory(self, inv, f):
        """Write inventory to a file.

        Note: this is a *whole inventory* operation, and should only be used
        sparingly, as it does not scale well with large trees.
        """
        raise NotImplementedError(self.write_inventory)

    def write_inventory_to_chunks(self, inv):
        """Produce a simple bytestring chunk representation of an inventory.

        Note: this is a *whole inventory* operation, and should only be used
        sparingly, as it does not scale well with large trees.

        The requirement for the contents of the string is that it can be passed
        to read_inventory_from_lines and the result is an identical inventory
        in memory.
        """
        raise NotImplementedError(self.write_inventory_to_chunks)

    def write_inventory_to_lines(self, inv):
        """Produce a simple lines representation of an inventory.

        Note: this is a *whole inventory* operation, and should only be used
        sparingly, as it does not scale well with large trees.

        The requirement for the contents of the string is that it can be passed
        to read_inventory_from_lines and the result is an identical inventory
        in memory.
        """
        raise NotImplementedError(self.write_inventory_to_lines)

    def read_inventory_from_lines(
        self, lines, revision_id=None, entry_cache=None, return_from_cache=False
    ):
        """Read bytestring chunks into an inventory object.

        :param lines: The serialized inventory to read.
        :param revision_id: If not-None, the expected revision id of the
            inventory. Some serialisers use this to set the results' root
            revision. This should be supplied for deserialising all
            from-repository inventories so that xml5 inventories that were
            serialised without a revision identifier can be given the right
            revision id (but not for working tree inventories where users can
            edit the data without triggering checksum errors or anything).
        :param entry_cache: An optional cache of InventoryEntry objects. If
            supplied we will look up entries via (file_id, revision_id) which
            should map to a valid InventoryEntry (File/Directory/etc) object.
        :param return_from_cache: Return entries directly from the cache,
            rather than copying them first. This is only safe if the caller
            promises not to mutate the returned inventory entries, but it can
            make some operations significantly faster.
        """
        raise NotImplementedError(self.read_inventory_from_lines)

    def read_inventory(self, f, revision_id=None):
        """See read_inventory_from_lines."""
        raise NotImplementedError(self.read_inventory)


class SerializerRegistry(registry.Registry):
    """Registry for serializer objects."""


revision_format_registry = SerializerRegistry()
revision_format_registry.register_lazy(
    "5", "bzrformats._bzr_rs", "revision_serializer_v5"
)
revision_format_registry.register_lazy(
    "8", "bzrformats._bzr_rs", "revision_serializer_v8"
)
revision_format_registry.register_lazy(
    "10", "bzrformats._bzr_rs", "revision_bencode_serializer"
)


inventory_format_registry = SerializerRegistry()
inventory_format_registry.register_lazy(
    "5", "bzrformats.xml5", "inventory_serializer_v5"
)
inventory_format_registry.register_lazy(
    "6", "bzrformats.xml6", "inventory_serializer_v6"
)
inventory_format_registry.register_lazy(
    "7", "bzrformats.xml7", "inventory_serializer_v7"
)
inventory_format_registry.register_lazy(
    "8", "bzrformats.xml8", "inventory_serializer_v8"
)
inventory_format_registry.register_lazy(
    "9", "bzrformats.chk_serializer", "inventory_chk_serializer_255_bigpage_9"
)
inventory_format_registry.register_lazy(
    "10", "bzrformats.chk_serializer", "inventory_chk_serializer_255_bigpage_10"
)



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for bzrformats osutils."""

import hashlib
import os

from .. import osutils
from . import TestCase, TestCaseInTempDir


class TestShaFunctions(TestCase):
    """Test the sha_string and sha_strings functions."""

    def test_sha_string_bytes(self):
        """Test sha_string with bytes input."""
        result = osutils.sha_string(b"hello world")
        expected = hashlib.sha1(b"hello world").hexdigest().encode("ascii")  # noqa: S324
        self.assertEqual(expected, result)

    def test_sha_string_unicode(self):
        """Test sha_string with unicode input."""
        result = osutils.sha_string("hello world")
        expected = hashlib.sha1(b"hello world").hexdigest().encode("ascii")  # noqa: S324
        self.assertEqual(expected, result)

    def test_sha_strings(self):
        """Test sha_strings with mixed input."""
        result = osutils.sha_strings([b"hello", " ", "world"])
        sha = hashlib.sha1()  # noqa: S324
        sha.update(b"hello")
        sha.update(b" ")
        sha.update(b"world")
        expected = sha.hexdigest().encode("ascii")
        self.assertEqual(expected, result)


class TestOsutilsFunctions(TestCase):
    """Test various osutils functions."""

    def test_split_unicode(self):
        """Test split with unicode paths."""
        dirname, basename = osutils.split("foo/bar")
        self.assertEqual("foo", dirname)
        self.assertEqual("bar", basename)

    def test_split_bytes(self):
        """Test split with byte paths."""
        dirname, basename = osutils.split(b"foo/bar")
        self.assertEqual(b"foo", dirname)
        self.assertEqual(b"bar", basename)

    def test_pathjoin_unicode(self):
        """Test pathjoin with unicode paths."""
        result = osutils.pathjoin("foo", "bar", "baz")
        self.assertEqual(os.path.join("foo", "bar", "baz"), result)

    def test_pathjoin_bytes(self):
        """Test pathjoin with byte paths."""
        result = osutils.pathjoin(b"foo", b"bar", b"baz")
        self.assertEqual(os.path.join(b"foo", b"bar", b"baz"), result)

    def test_basename_unicode(self):
        """Test basename with unicode path."""
        result = osutils.basename("foo/bar/baz")
        self.assertEqual("baz", result)

    def test_basename_bytes(self):
        """Test basename with byte path."""
        result = osutils.basename(b"foo/bar/baz")
        self.assertEqual(b"baz", result)

    def test_dirname_unicode(self):
        """Test dirname with unicode path."""
        result = osutils.dirname("foo/bar/baz")
        self.assertEqual("foo/bar", result)

    def test_dirname_bytes(self):
        """Test dirname with byte path."""
        result = osutils.dirname(b"foo/bar/baz")
        self.assertEqual(b"foo/bar", result)

    def test_splitpath(self):
        """Test splitpath function."""
        self.assertEqual(["foo", "bar"], osutils.splitpath("foo/bar"))
        self.assertEqual(["foo", "bar"], osutils.splitpath("/foo/bar"))
        self.assertEqual([b"foo", b"bar"], osutils.splitpath(b"foo/bar"))
        self.assertEqual([b"foo", b"bar"], osutils.splitpath(b"/foo/bar"))
        self.assertEqual([], osutils.splitpath(""))
        self.assertEqual([], osutils.splitpath("/"))

    def test_contains_whitespace(self):
        """Test contains_whitespace function."""
        self.assertTrue(osutils.contains_whitespace("hello world"))
        self.assertTrue(osutils.contains_whitespace("hello\tworld"))
        self.assertTrue(osutils.contains_whitespace("hello\nworld"))
        self.assertFalse(osutils.contains_whitespace("helloworld"))

        # Test bytes
        self.assertTrue(osutils.contains_whitespace(b"hello world"))
        self.assertFalse(osutils.contains_whitespace(b"helloworld"))

    def test_normalized_filename(self):
        """Test normalized_filename function."""
        # Simple ASCII filename
        result, can_access = osutils.normalized_filename("test.txt")
        self.assertEqual("test.txt", result)
        self.assertTrue(can_access)

        # Bytes filename
        result, can_access = osutils.normalized_filename(b"test.txt")
        self.assertEqual(b"test.txt", result)
        self.assertTrue(can_access)

    def test_chunks_to_lines(self):
        """Test chunks_to_lines function."""
        chunks = [b"line1\n", b"line2\nli", b"ne3\n"]
        result = osutils.chunks_to_lines(chunks)
        self.assertEqual([b"line1\n", b"line2\n", b"line3\n"], result)

        # Test with no newline at end
        chunks = [b"line1\n", b"line2"]
        result = osutils.chunks_to_lines(chunks)
        self.assertEqual([b"line1\n", b"line2"], result)

        # Test empty chunks
        self.assertEqual([], osutils.chunks_to_lines([]))

    def test_chunks_to_lines_iter(self):
        """Test chunks_to_lines_iter function."""
        chunks = iter([b"line1\n", b"line2\nli", b"ne3\n"])
        result = list(osutils.chunks_to_lines_iter(chunks))
        self.assertEqual([b"line1\n", b"line2\n", b"line3\n"], result)


class TestRustOsutilsFunctions(TestCase):
    """Test the Rust-based osutils functions."""

    def test_rand_chars(self):
        """Test rand_chars generates the right length string."""
        result = osutils.rand_chars(10)
        self.assertEqual(10, len(result))
        # Should only contain alphanumeric characters
        self.assertTrue(all(c.isalnum() for c in result))

    def test_is_inside(self):
        """Test is_inside function."""
        # Should work with both strings and bytes
        self.assertTrue(osutils.is_inside("/home", "/home/user"))
        self.assertTrue(osutils.is_inside("/home/", "/home/user"))
        self.assertFalse(osutils.is_inside("/home", "/usr/bin"))
        self.assertFalse(osutils.is_inside("/home/user", "/home"))

    def test_is_inside_any(self):
        """Test is_inside_any function."""
        dirs = ["/home", "/usr"]
        self.assertTrue(osutils.is_inside_any(dirs, "/home/user"))
        self.assertTrue(osutils.is_inside_any(dirs, "/usr/bin"))
        self.assertFalse(osutils.is_inside_any(dirs, "/var/log"))

    def test_parent_directories(self):
        """Test parent_directories function."""
        result = osutils.parent_directories("/home/user/documents/file.txt")
        # Convert to list since it returns an iterator
        parents = list(result)
        self.assertIn("/home/user/documents", parents)
        self.assertIn("/home/user", parents)
        self.assertIn("/home", parents)


class TestFileIterator(TestCase):
    """Test file_iterator function."""

    def test_file_iterator(self):
        """Test iterating over file contents."""
        import io

        content = b"a" * 100000  # 100KB of data
        file_obj = io.BytesIO(content)

        chunks = list(osutils.file_iterator(file_obj, chunk_size=1024))

        # Should have multiple chunks
        self.assertTrue(len(chunks) > 1)

        # Reassemble and check
        reassembled = b"".join(chunks)
        self.assertEqual(content, reassembled)

        # Check chunk sizes (all but last should be 1024)
        for chunk in chunks[:-1]:
            self.assertEqual(1024, len(chunk))


class TestPumpfile(TestCaseInTempDir):
    """Test pumpfile function."""

    def test_pumpfile(self):
        """Test copying data between file objects."""
        import io

        # Create source with some data
        source_data = b"Hello, world!" * 1000
        source = io.BytesIO(source_data)

        # Create destination
        dest = io.BytesIO()

        # Pump the data
        bytes_copied = osutils.pumpfile(source, dest)

        # Check the result
        self.assertEqual(len(source_data), bytes_copied)
        self.assertEqual(source_data, dest.getvalue())


class TestFileKindFromStatMode(TestCase):
    """Test file_kind_from_stat_mode function."""

    def test_regular_file(self):
        """Test regular file detection."""
        import stat

        mode = stat.S_IFREG | 0o644
        self.assertEqual("file", osutils.file_kind_from_stat_mode(mode))

    def test_directory(self):
        """Test directory detection."""
        import stat

        mode = stat.S_IFDIR | 0o755
        self.assertEqual("directory", osutils.file_kind_from_stat_mode(mode))

    def test_symlink(self):
        """Test symlink detection."""
        import stat

        mode = stat.S_IFLNK | 0o777
        self.assertEqual("symlink", osutils.file_kind_from_stat_mode(mode))

    def test_fifo(self):
        """Test FIFO detection."""
        import stat

        mode = stat.S_IFIFO | 0o666
        self.assertEqual("fifo", osutils.file_kind_from_stat_mode(mode))

    def test_socket(self):
        """Test socket detection."""
        import stat

        mode = stat.S_IFSOCK | 0o666
        self.assertEqual("socket", osutils.file_kind_from_stat_mode(mode))

    def test_char_device(self):
        """Test character device detection."""
        import stat

        mode = stat.S_IFCHR | 0o666
        self.assertEqual("chardev", osutils.file_kind_from_stat_mode(mode))

    def test_block_device(self):
        """Test block device detection."""
        import stat

        mode = stat.S_IFBLK | 0o666
        self.assertEqual("block", osutils.file_kind_from_stat_mode(mode))


# Add test module discovery
def test_suite():
    """Return the test suite for osutils tests."""
    import unittest

    return unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])


if __name__ == "__main__":
    import unittest

    unittest.main()



# Copyright (C) 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Inventory delta serialisation.

See doc/developers/inventory.txt for the description of the format.

In this module the interesting classes are:
 - InventoryDeltaSerializer - object to read/write inventory deltas.
"""

__all__ = ["InventoryDeltaSerializer"]

from ._bzr_rs import inventory as _inventory_delta_rs
from .revision import RevisionID

InventoryDeltaError = _inventory_delta_rs.InventoryDeltaError
IncompatibleInventoryDelta = _inventory_delta_rs.IncompatibleInventoryDelta
parse_inventory_entry = _inventory_delta_rs.parse_inventory_entry
serialize_inventory_entry = _inventory_delta_rs.serialize_inventory_entry
InventoryDelta = _inventory_delta_rs.InventoryDelta


class InventoryDeltaSerializer:
    """Serialize inventory deltas."""

    def __init__(self, versioned_root, tree_references):
        """Create an InventoryDeltaSerializer.

        :param versioned_root: If True, any root entry that is seen is expected
            to be versioned, and root entries can have any fileid.
        :param tree_references: If True support tree-reference entries.
        """
        self._versioned_root = versioned_root
        self._tree_references = tree_references

    def delta_to_lines(
        self,
        old_name: RevisionID,
        new_name: RevisionID,
        delta_to_new: _inventory_delta_rs.InventoryDelta,
    ):
        """Return a line sequence for delta_to_new.

        Both the versioned_root and tree_references flags must be set via
        require_flags before calling this.

        :param old_name: A UTF8 revision id for the old inventory.  May be
            NULL_REVISION if there is no older inventory and delta_to_new
            includes the entire inventory contents.
        :param new_name: The version name of the inventory we create with this
            delta.
        :param delta_to_new: An inventory delta such as Inventory.apply_delta
            takes.
        :return: The serialized delta as lines.
        """
        return _inventory_delta_rs.serialize_inventory_delta(
            old_name,
            new_name,
            delta_to_new,
            self._versioned_root,
            self._tree_references,
        )


class InventoryDeltaDeserializer:
    """Deserialize inventory deltas."""

    def __init__(self, allow_versioned_root=True, allow_tree_references=True):
        """Create an InventoryDeltaDeserializer.

        :param versioned_root: If True, any root entry that is seen is expected
            to be versioned, and root entries can have any fileid.
        :param tree_references: If True support tree-reference entries.
        """
        self._allow_versioned_root = allow_versioned_root
        self._allow_tree_references = allow_tree_references

    def parse_text_bytes(self, lines):
        """Parse the text bytes of a serialized inventory delta.

        If versioned_root and/or tree_references flags were set via
        require_flags, then the parsed flags must match or a BzrError will be
        raised.

        :param lines: The lines to parse. This can be obtained by calling
            delta_to_lines.
        :return: (parent_id, new_id, versioned_root, tree_references,
            inventory_delta)
        """
        return _inventory_delta_rs.parse_inventory_delta(
            lines, self._allow_versioned_root, self._allow_tree_references
        )



# Copyright (C) 2008-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the python and pyrex extensions of groupcompress."""

import sys

from testscenarios import load_tests_apply_scenarios

from .. import groupcompress
from .._bzr_rs import groupcompress as _groupcompress_rs
from . import TestCase, _try_import

_compiled_groupcompress_module = _try_import("bzrformats._groupcompress_pyx")


def module_scenarios():
    scenarios = [
        (
            "line",
            {"make_delta": groupcompress.make_line_delta},
        ),
        ("rabin", {"make_delta": groupcompress.make_rabin_delta}),
    ]
    return scenarios


def two_way_scenarios():
    scenarios = [
        ("LR", {"make_delta": groupcompress.make_line_delta}),
        ("RR", {"make_delta": groupcompress.make_rabin_delta}),
    ]
    return scenarios


load_tests = load_tests_apply_scenarios


try:
    from bzrformats import _groupcompress_pyx as _groupcompress_cython
except ImportError:
    _groupcompress_cython = None

_text1 = b"""\
This is a bit
of source text
which is meant to be matched
against other text
"""

_text2 = b"""\
This is a bit
of source text
which is meant to differ from
against other text
"""

_text3 = b"""\
This is a bit
of source text
which is meant to be matched
against other text
except it also
has a lot more data
at the end of the file
"""

_first_text = b"""\
a bit of text, that
does not have much in
common with the next text
"""

_second_text = b"""\
some more bit of text, that
does not have much in
common with the previous text
and has some extra text
"""


_third_text = b"""\
a bit of text, that
has some in common with the previous text
and has some extra text
and not have much in
common with the next text
"""

_fourth_text = b"""\
123456789012345
same rabin hash
123456789012345
same rabin hash
123456789012345
same rabin hash
123456789012345
same rabin hash
"""


class TestMakeAndApplyDelta(TestCase):
    scenarios = module_scenarios()
    _gc_module = None  # Set by load_tests

    def setUp(self):
        super().setUp()
        self.apply_delta = _groupcompress_rs.apply_delta
        self.apply_delta_to_source = _groupcompress_rs.apply_delta_to_source

    def test_make_delta_is_typesafe(self):
        self.make_delta(b"a string", b"another string")

        def _check_make_delta(string1, string2):
            self.assertRaises(TypeError, self.make_delta, string1, string2)

        _check_make_delta(b"a string", object())
        _check_make_delta(b"a string", "not a string")
        _check_make_delta(object(), b"a string")
        _check_make_delta("not a string", b"a string")

    def test_make_noop_delta(self):
        ident_delta = self.make_delta(_text1, _text1)
        self.assertEqual(b"M\x90M", ident_delta)
        ident_delta = self.make_delta(_text2, _text2)
        self.assertEqual(b"N\x90N", ident_delta)
        ident_delta = self.make_delta(_text3, _text3)
        self.assertEqual(b"\x87\x01\x90\x87", ident_delta)

    def assertDeltaIn(self, delta1, delta2, delta):
        """Make sure that the delta bytes match one of the expectations."""
        # In general, the python delta matcher gives different results than the
        # pyrex delta matcher. Both should be valid deltas, though.
        if delta not in (delta1, delta2):
            self.fail(
                b"Delta bytes:\n"
                b"       %r\n"
                b"not in %r\n"
                b"    or %r" % (delta, delta1, delta2)
            )

    def test_make_delta(self):
        delta = self.make_delta(_text1, _text2)
        self.assertDeltaIn(
            b"N\x90/\x1fdiffer from\nagainst other text\n",
            b"N\x90\x1d\x1ewhich is meant to differ from\n\x91:\x13",
            delta,
        )
        delta = self.make_delta(_text2, _text1)
        self.assertDeltaIn(
            b"M\x90/\x1ebe matched\nagainst other text\n",
            b"M\x90\x1d\x1dwhich is meant to be matched\n\x91;\x13",
            delta,
        )
        delta = self.make_delta(_text3, _text1)
        self.assertEqual(b"M\x90M", delta)
        delta = self.make_delta(_text3, _text2)
        self.assertDeltaIn(
            b"N\x90/\x1fdiffer from\nagainst other text\n",
            b"N\x90\x1d\x1ewhich is meant to differ from\n\x91:\x13",
            delta,
        )

    def test_make_delta_with_large_copies(self):
        # We want to have a copy that is larger than 64kB, which forces us to
        # issue multiple copy instructions.
        big_text = _text3 * 1220
        delta = self.make_delta(big_text, big_text)
        self.assertDeltaIn(
            b"\xdc\x86\x0a"  # Encoding the length of the uncompressed text
            b"\x80"  # Copy 64kB, starting at byte 0
            b"\x84\x01"  # and another 64kB starting at 64kB
            b"\xb4\x02\x5c\x83",  # And the bit of tail.
            None,  # Both implementations should be identical
            delta,
        )

    def test_apply_delta_is_typesafe(self):
        self.apply_delta(_text1, b"M\x90M")
        self.assertRaises(TypeError, self.apply_delta, object(), b"M\x90M")
        self.assertRaises(
            (ValueError, TypeError),
            self.apply_delta,
            _text1.decode("latin1"),
            b"M\x90M",
        )
        self.assertRaises((ValueError, TypeError), self.apply_delta, _text1, "M\x90M")
        self.assertRaises(TypeError, self.apply_delta, _text1, object())

    def test_apply_delta(self):
        target = self.apply_delta(
            _text1, b"N\x90/\x1fdiffer from\nagainst other text\n"
        )
        self.assertEqual(_text2, target)
        target = self.apply_delta(_text2, b"M\x90/\x1ebe matched\nagainst other text\n")
        self.assertEqual(_text1, target)

    def test_apply_delta_to_source_is_safe(self):
        self.assertRaises(TypeError, self.apply_delta_to_source, object(), 0, 1)
        self.assertRaises(TypeError, self.apply_delta_to_source, "unicode str", 0, 1)
        # end > length
        self.assertRaises(ValueError, self.apply_delta_to_source, b"foo", 1, 4)
        # start > length
        self.assertRaises(ValueError, self.apply_delta_to_source, b"foo", 5, 3)
        # start > end
        self.assertRaises(ValueError, self.apply_delta_to_source, b"foo", 3, 2)

    def test_apply_delta_to_source(self):
        source_and_delta = _text1 + b"N\x90/\x1fdiffer from\nagainst other text\n"
        self.assertEqual(
            _text2,
            self.apply_delta_to_source(
                source_and_delta, len(_text1), len(source_and_delta)
            ),
        )


class TestMakeAndApplyCompatible(TestCase):
    scenarios = two_way_scenarios()

    make_delta = None  # Set by load_tests
    apply_delta = _groupcompress_rs.apply_delta

    def assertMakeAndApply(self, source, target):
        """Assert that generating a delta and applying gives success."""
        delta = self.make_delta(source, target)
        bytes = self.apply_delta(source, delta)
        self.assertEqualDiff(target, bytes)

    def test_direct(self):
        self.assertMakeAndApply(_text1, _text2)
        self.assertMakeAndApply(_text2, _text1)
        self.assertMakeAndApply(_text1, _text3)
        self.assertMakeAndApply(_text3, _text1)
        self.assertMakeAndApply(_text2, _text3)
        self.assertMakeAndApply(_text3, _text2)


class TestDeltaIndex(TestCase):
    def setUp(self):
        super().setUp()
        # This test isn't multiplied, because we only have DeltaIndex for the
        # compiled form
        # We call this here, because _test_needs_features happens after setUp
        if _groupcompress_cython is None:
            self.skipTest("Cython _groupcompress module not available")
        self._gc_module = _groupcompress_cython

    def test_repr(self):
        di = self._gc_module.DeltaIndex(b"test text\n")
        self.assertEqual("DeltaIndex(1, 10)", repr(di))

    def test_sizeof(self):
        di = self._gc_module.DeltaIndex()
        # Exact value will depend on platform but should include sources
        # source_info is a pointer and two longs so at least 12 bytes
        lower_bound = di._max_num_sources * 12
        self.assertGreater(sys.getsizeof(di), lower_bound)

    def test__dump_no_index(self):
        di = self._gc_module.DeltaIndex()
        self.assertEqual(None, di._dump_index())

    def test__dump_index_simple(self):
        di = self._gc_module.DeltaIndex()
        di.add_source(_text1, 0)
        self.assertFalse(di._has_index())
        self.assertEqual(None, di._dump_index())
        _ = di.make_delta(_text1)
        self.assertTrue(di._has_index())
        hash_list, entry_list = di._dump_index()
        self.assertEqual(16, len(hash_list))
        self.assertEqual(68, len(entry_list))
        just_entries = [
            (idx, text_offset, hash_val)
            for idx, (text_offset, hash_val) in enumerate(entry_list)
            if text_offset != 0 or hash_val != 0
        ]
        rabin_hash = groupcompress.rabin_hash
        self.assertEqual(
            [
                (8, 16, rabin_hash(_text1[1:17])),
                (25, 48, rabin_hash(_text1[33:49])),
                (34, 32, rabin_hash(_text1[17:33])),
                (47, 64, rabin_hash(_text1[49:65])),
            ],
            just_entries,
        )
        # This ensures that the hash map points to the location we expect it to
        for entry_idx, _text_offset, hash_val in just_entries:
            self.assertEqual(entry_idx, hash_list[hash_val & 0xF])

    def test__dump_index_two_sources(self):
        di = self._gc_module.DeltaIndex()
        di.add_source(_text1, 0)
        di.add_source(_text2, 2)
        start2 = len(_text1) + 2
        self.assertTrue(di._has_index())
        hash_list, entry_list = di._dump_index()
        self.assertEqual(16, len(hash_list))
        self.assertEqual(68, len(entry_list))
        just_entries = [
            (idx, text_offset, hash_val)
            for idx, (text_offset, hash_val) in enumerate(entry_list)
            if text_offset != 0 or hash_val != 0
        ]
        rabin_hash = groupcompress.rabin_hash
        self.assertEqual(
            [
                (8, 16, rabin_hash(_text1[1:17])),
                (9, start2 + 16, rabin_hash(_text2[1:17])),
                (25, 48, rabin_hash(_text1[33:49])),
                (30, start2 + 64, rabin_hash(_text2[49:65])),
                (34, 32, rabin_hash(_text1[17:33])),
                (35, start2 + 32, rabin_hash(_text2[17:33])),
                (43, start2 + 48, rabin_hash(_text2[33:49])),
                (47, 64, rabin_hash(_text1[49:65])),
            ],
            just_entries,
        )
        # Each entry should be in the appropriate hash bucket.
        for entry_idx, _text_offset, hash_val in just_entries:
            hash_idx = hash_val & 0xF
            self.assertTrue(hash_list[hash_idx] <= entry_idx < hash_list[hash_idx + 1])

    def test_first_add_source_doesnt_index_until_make_delta(self):
        di = self._gc_module.DeltaIndex()
        self.assertFalse(di._has_index())
        di.add_source(_text1, 0)
        self.assertFalse(di._has_index())
        # However, asking to make a delta will trigger the index to be
        # generated, and will generate a proper delta
        delta = di.make_delta(_text2)
        self.assertTrue(di._has_index())
        self.assertEqual(b"N\x90/\x1fdiffer from\nagainst other text\n", delta)

    def test_add_source_max_bytes_to_index(self):
        di = self._gc_module.DeltaIndex()
        di._max_bytes_to_index = 3 * 16
        di.add_source(_text1, 0)  # (77 bytes -1) // 3 = 25 byte stride
        di.add_source(_text3, 3)  # (135 bytes -1) // 3 = 44 byte stride
        start2 = len(_text1) + 3
        hash_list, entry_list = di._dump_index()
        self.assertEqual(16, len(hash_list))
        self.assertEqual(67, len(entry_list))
        just_entries = sorted(
            [
                (text_offset, hash_val)
                for text_offset, hash_val in entry_list
                if text_offset != 0 or hash_val != 0
            ]
        )
        rabin_hash = groupcompress.rabin_hash
        self.assertEqual(
            [
                (25, rabin_hash(_text1[10:26])),
                (50, rabin_hash(_text1[35:51])),
                (75, rabin_hash(_text1[60:76])),
                (start2 + 44, rabin_hash(_text3[29:45])),
                (start2 + 88, rabin_hash(_text3[73:89])),
                (start2 + 132, rabin_hash(_text3[117:133])),
            ],
            just_entries,
        )

    def test_second_add_source_triggers_make_index(self):
        di = self._gc_module.DeltaIndex()
        self.assertFalse(di._has_index())
        di.add_source(_text1, 0)
        self.assertFalse(di._has_index())
        di.add_source(_text2, 0)
        self.assertTrue(di._has_index())

    def test_make_delta(self):
        di = self._gc_module.DeltaIndex(_text1)
        delta = di.make_delta(_text2)
        self.assertEqual(b"N\x90/\x1fdiffer from\nagainst other text\n", delta)

    def test_delta_against_multiple_sources(self):
        di = self._gc_module.DeltaIndex()
        di.add_source(_first_text, 0)
        self.assertEqual(len(_first_text), di._source_offset)
        di.add_source(_second_text, 0)
        self.assertEqual(len(_first_text) + len(_second_text), di._source_offset)
        delta = di.make_delta(_third_text)
        result = _groupcompress_rs.apply_delta(_first_text + _second_text, delta)
        self.assertEqualDiff(_third_text, result)
        self.assertEqual(
            b'\x85\x01\x90\x14\x0chas some in \x91v6\x03and\x91d"\x91:\n', delta
        )

    def test_delta_with_offsets(self):
        di = self._gc_module.DeltaIndex()
        di.add_source(_first_text, 5)
        self.assertEqual(len(_first_text) + 5, di._source_offset)
        di.add_source(_second_text, 10)
        self.assertEqual(len(_first_text) + len(_second_text) + 15, di._source_offset)
        delta = di.make_delta(_third_text)
        self.assertIsNot(None, delta)
        result = _groupcompress_rs.apply_delta(
            b"12345" + _first_text + b"1234567890" + _second_text, delta
        )
        self.assertIsNot(None, result)
        self.assertEqualDiff(_third_text, result)
        self.assertEqual(
            b'\x85\x01\x91\x05\x14\x0chas some in \x91\x856\x03and\x91s"\x91?\n',
            delta,
        )

    def test_delta_with_delta_bytes(self):
        di = self._gc_module.DeltaIndex()
        source = _first_text
        di.add_source(_first_text, 0)
        self.assertEqual(len(_first_text), di._source_offset)
        delta = di.make_delta(_second_text)
        self.assertEqual(
            b"h\tsome more\x91\x019&previous text\nand has some extra text\n", delta
        )
        di.add_delta_source(delta, 0)
        source += delta
        self.assertEqual(len(_first_text) + len(delta), di._source_offset)
        second_delta = di.make_delta(_third_text)
        result = _groupcompress_rs.apply_delta(source, second_delta)
        self.assertEqualDiff(_third_text, result)
        # We should be able to match against the
        # 'previous text\nand has some...'  that was part of the delta bytes
        # Note that we don't match the 'common with the', because it isn't long
        # enough to match in the original text, and those bytes are not present
        # in the delta for the second text.
        self.assertEqual(
            b"\x85\x01\x90\x14\x1chas some in common with the \x91S&\x03and\x91\x18,",
            second_delta,
        )
        # Add this delta, and create a new delta for the same text. We should
        # find the remaining text, and only insert the short 'and' text.
        di.add_delta_source(second_delta, 0)
        source += second_delta
        third_delta = di.make_delta(_third_text)
        result = _groupcompress_rs.apply_delta(source, third_delta)
        self.assertEqualDiff(_third_text, result)
        self.assertEqual(
            b"\x85\x01\x90\x14\x91\x7e\x1c\x91S&\x03and\x91\x18,", third_delta
        )
        # Now create a delta, which we know won't be able to be 'fit' into the
        # existing index
        fourth_delta = di.make_delta(_fourth_text)
        self.assertEqual(
            _fourth_text, _groupcompress_rs.apply_delta(source, fourth_delta)
        )
        self.assertEqual(
            b"\x80\x01"
            b"\x7f123456789012345\nsame rabin hash\n"
            b"123456789012345\nsame rabin hash\n"
            b"123456789012345\nsame rabin hash\n"
            b"123456789012345\nsame rabin hash"
            b"\x01\n",
            fourth_delta,
        )
        di.add_delta_source(fourth_delta, 0)
        source += fourth_delta
        # With the next delta, everything should be found
        fifth_delta = di.make_delta(_fourth_text)
        self.assertEqual(
            _fourth_text, _groupcompress_rs.apply_delta(source, fifth_delta)
        )
        self.assertEqual(b"\x80\x01\x91\xa7\x7f\x01\n", fifth_delta)



# Copyright (C) 2007-2010 Canonical Ltd
# Copyright (C) 2018 Breezy developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# cython: language_level=3

"""Trivial string helpers for use in other cython modules."""

cdef extern from "python-compat.h":
    object PyBytes_FromStringAndSize (char *, Py_ssize_t)


cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise
    # memrchr seems to be a GNU extension, so we have to implement it ourselves
    cdef char *pos
    cdef char *start

    start = <char*>s
    pos = start + n - 1
    while pos >= start:
        if pos[0] == c:
            return <void*>pos
        pos = pos - 1
    return NULL


cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
    if size < 0:
        raise AssertionError(
            'tried to create a string with an invalid size: %d' % size)
    return PyBytes_FromStringAndSize(s, size)


cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
    if size < 0:
        raise AssertionError(
            'tried to create a string with an invalid size: %d' % size)
    # For now, don't intern on Python 3
    return PyBytes_FromStringAndSize(s, size)
